/* $Id: export.c,v 1.30 2002/12/18 22:12:18 classic Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    export: Exports Database in Repligard format
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <zlib.h>
#include "midgard/midgard.h"
#include "schema.h"
#include "rpconfig.h"
#include "common.h"
#include "export.h"
#include "exporthook.h"

static void
rpd_write_object_field (gpointer key, gpointer value, gpointer user_data)
{
  WObjectFD *ofd = user_data;
  RSchemaField *field = value;
  WGuidCacheEntry *entry;
  const gchar *str;
  const gchar *c;
  gchar *link;
  int len;
  gchar tempname[255], temptable[255];

  str = mgd_column (ofd->res, (gchar *) key);
  if (!str)
    return;
  len = strlen (str);
  /* We are writing record's fields in a form of 
     <name>
     data
     </name>
     where data representation depends on its nature.
   */
  if (len != 0)
    {
/*	        if (configuration.verbose)
                             g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "got field type %s", 
			    		 key );
*/
      if (field->type == R_BLOB_FIELD)
	gzprintf (ofd->fd, "\t<%s locid=\"%s\">", (gchar *) key, str);
      else
	gzprintf (ofd->fd, "\t<%s>", (gchar *) key);
      switch (field->type)
	{
	case R_INTEGER_FIELD:
	case R_STRING_FIELD:
	case R_PASSWORD_FIELD:
	  /* INTEGERs, STRINGs and PASSWORDs use raw dumping for the moment */
	  if (!
	      (configuration.nopasswords &&
	      ((!g_strcasecmp((gchar *) key, "username"))
	     || (!g_strcasecmp((gchar *) key, "password")))))
  	     gzputs (ofd->fd, str);
	  break;
	case R_TEXT_FIELD:
	  if (!
	      (configuration.nopasswords &&
	      (!g_strcasecmp((gchar *) key, "username"))))  {
            if (configuration.quote_as_cdata) {
               gzputs (ofd->fd, "<![CDATA[");
               gzputs (ofd->fd, str);
               gzputs (ofd->fd, "]]>");
            } else {
               for (c = str; *c != '\0'; c++) {
                  if (*c == '<')          {  gzputs (ofd->fd, "&lt;"); }
                  else if (*c == '>')     {  gzputs (ofd->fd, "&gt;"); }
                  else if (*c == '&')     {  gzputs (ofd->fd, "&amp;"); }
                  else if (isascii(*c))   {  gzputc (ofd->fd, *c); }
                  else                    {  gzprintf (ofd->fd, "&x%X;", *c); }
               }
            }
	   }
	  break;
	case R_LINK_FIELD:
	case R_LINK_FIELD_NODEPS:
	  /* Special case: we are writing links only to existing objects.
	     For top-level objects which have no upper level links, this
	     field will be empty */
	  if (atoi (str) != 0)
	    {
	      link = g_strconcat (field->link, " ", str, NULL);
	      entry = ((WGuidCacheEntry *)
		       g_hash_table_lookup (ofd->guid_hash, link));
	      if (entry)
		{

		  gzputs (ofd->fd, entry->guid);
		  /* Check if the link is already replicated */
		  if (entry->replicated == FALSE)
		    {

		      RListEntry *depitem = g_new (RListEntry, 1);
		      /* Add this link into the list of items requested for replication by dependency */
		      depitem->guid = FALSE;
		      depitem->type = g_strdup (field->link);
		      depitem->id = g_strdup (str);
		      g_list_append (ofd->exp_data->dependencies, depitem);
		      if (configuration.verbose > 2)
			g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			       "LINK add dep: %s %s", depitem->type,
			       depitem->id);

		    }
		}
	      g_free (link);
	    }
	  break;
	case R_RLINK_FIELD:
	  /* Special case: this is the link from the BLOBS or RECORD_EXTENSION
	     so we need to make correct GUID by reusing the id encoded in the RLINK.
	   */
	 
	  if (atoi (str) != 0)
	    {
	      (void) sscanf (field->link, "%s %s", temptable, tempname);
	     
	      link =
		g_strconcat (mgd_column (ofd->res, temptable),
			     " ", str, NULL);
	      entry = ((WGuidCacheEntry *)
		       g_hash_table_lookup (ofd->guid_hash, link));
	      if (entry)
		{
		  gzputs (ofd->fd, entry->guid);
		  /* Check if the link is already replicated */
		  if (entry->replicated == FALSE)
		    {
		      RListEntry *depitem = g_new (RListEntry, 1);
		      /* Add this link into the list of items requested for replication by dependency */
		      depitem->guid = FALSE;
		      depitem->type =
			g_strdup (mgd_column (ofd->res, temptable));
		      depitem->id = g_strdup (str);
		      g_list_append (ofd->exp_data->dependencies, depitem);
		      if (configuration.verbose > 1)
			g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			       "adding reversed dependency: %s %s",
			       depitem->type, depitem->id);
		    }
		}
	      g_free (link);
	    }
	  break;
	case R_BLOB_FIELD:
	  /* Special case: this is the link to external BLOB container. We should
	     encode content of this container using BASE64 encoding and output it
	     instead of link itself. We don't use CDATA section because BASE64
	     is XML-safe.
	   */
	  gzputc (ofd->fd, '\n');
	  repligard_encode_blob (str, ofd->fd);
	  /* We've already put '\n' in the encoder, so put '\t' now
	     purely for beauty  :-)
	   */
	  gzputc (ofd->fd, '\t');
	  break;
	}
      gzprintf (ofd->fd, "</%s>\n", (gchar *) key);
    }
  else
    {
      gzprintf (ofd->fd, "\t<%s/>\n", (gchar *) key);
    }
}

/* Writes object passed by res into file (fd).
   table: resource name
   object_hash: resource description from schema
   res: resource fetched from database
   exp_data: export description block
*/
static void
repligard_write_object (GHashTable * object_hash, WExport * exp_data,
			const gchar * table, midgard_res * res)
{
  gchar *guid;
  WObjectFD ofd;
  WObjectType otype;
  WGuidCacheEntry *entry;
  gchar *usefrom;
  gint export_result;

  /* Special case for empty queries */
  if (!res)
    return;

  g_assert (exp_data != NULL);
  g_assert (table != NULL);
  g_assert (object_hash != NULL);

  ofd.guid_hash = exp_data->guid_hash;
  ofd.res = res;
  ofd.fd = exp_data->fd;
  ofd.exp_data = exp_data;
  guid = g_strconcat (table, " ", mgd_column (res, "id"), NULL);
  entry = g_hash_table_lookup (exp_data->guid_hash, guid);
  /* Write resource only if it's changed time is newer than updated one and
     this resource hasn't replicated yet
     We also write empty statement for deleted object to be able delete them
     at remote site.
   */
  if (entry)
    {
      if (entry->replicated == FALSE)
	{
	  if (configuration.verbose)
#if HAVE_MIDGARD_SITEGROUPS	  
    	    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "exporting resource %s '%s' from SG%s", 
		table, guid, mgd_column(res, "sitegroup"));
#else
	    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "exporting resource %s '%s'",
		table, guid);
#endif
          usefrom=  entry->updated;
 	  if (configuration.fromdate) usefrom = configuration.fromdate;
	      
	  if ((g_strcasecmp (entry->changed, usefrom ) > 0) || (configuration.diff == FALSE))
	    {
	      /* Export this object by default */
	      export_result = 0;
	      if (entry->deleted == TRUE)
		{
		  gzprintf (exp_data->fd,
			    "<%s id=\"%s\" deleted=\"%s\"/>\n",
			    table, entry->guid, entry->changed);
		}
	      else
		{
		  /* If export hook is set then run object through it */
		  if(configuration.export_hook != NULL) {
		    otype.res = res;
		    otype.otype = object_hash;
		    otype.guid = guid;
		    otype.ntype = table;
		    export_result = repligard_run_export_hook(&otype);
		  }
		  /* Export object only if export hook was successful */
		  if(!export_result) {
		    gzprintf (exp_data->fd,
			      "<%s id=\"%s\" changed=\"%s\">\n",
			      table, entry->guid, entry->changed);
		    g_hash_table_foreach (object_hash,
					  rpd_write_object_field, &ofd);
		    gzprintf (exp_data->fd, "</%s>\n", table);
		  }
		}
		/* Set updated time to changed one to avoid cross-referenced export */
		g_free (entry->updated);
		entry->updated = g_strdup (entry->changed);
		/* Note that we set 'replicated' in any case even if actual replication failed */
		/* This is perfectly understandable: export hook may forbid export from the _application_ */
		/* point of view, but from the Repligard point of view this object already participated in */
		/* replication so there is no need to try replicate it again */
		entry->replicated = TRUE;

	    }
	}
    }
  g_free (guid);

}

static void
rpd_write_links_foreach (gpointer key, gpointer value, gpointer user_data)
{
  struct
  {
    WExport *exp_data;
    int *tree;
  }
   *pt = user_data;
  RSchemaField *val = value;
  gchar temptable[255], tempname[255], temptablename[255], tempidcol[255];

  if (!pt)
    return;
  if (configuration.verbose> 1) g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "  LINKS FOREACH %d", val->type);
	     
  if (pt->tree)
    {
      int *cycle = pt->tree;
      if (val->type == R_RLINK_FIELD)
	{
	  (void) sscanf (key, "%s %s %s %s", temptable, tempname,
			 temptablename, tempidcol);
	}
      else
	{
	  (void) sscanf (key, "%s %s", temptable, tempname);
	}
      while (cycle && (*cycle > 0))
	{
	  RListEntry *depitem;
	  midgard_res *res;
	  res=NULL;
	  if (val->type == R_RLINK_FIELD)
	    {
	       if (configuration.verbose> 1)
	         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "      dependency look: %d %s %s %s %d %s %s", 
	         val->type,   val->link,   temptable, tempname, *cycle, temptablename, tempidcol);                                            
	        
	      /* NEEDS TO MAKE
	         SELECT id FROM val->link WHERE temptablename = tempname AND tempidcol = *cycle
	       */
	         res = mgd_sitegroup_select (pt->exp_data->mgd, "id",
					  val->link, "$s=$q AND $s=$d", NULL,
					  temptablename, tempname, temptable,
					  *cycle);
		 			    
	    }
	  else
	    {
	     if (val->type == R_LINK_FIELD) {
	     
	      res =
		mgd_sitegroup_select (pt->exp_data->mgd, "id",
				      val->link, "$s=$d", NULL,
				      temptable, *cycle);
              
	       if (configuration.verbose > 1)
	         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "      GETTING DEPS:%d from  %s where %s %d", 
	           val->type,  val->link,   temptable, *cycle);                                         
	          
	      }  
	    }
	  if (res)
	    {
	      while (mgd_fetch (res))
		if(pt->exp_data->expandtree == TRUE) {
		  /* Add this link into the list of items requested for replication by dependency */
		  depitem = g_new (RListEntry, 1);
		  depitem->guid = FALSE;
		  depitem->type = g_strdup (val->link);
		  depitem->id = g_strdup (mgd_column (res, "id"));
		  g_list_append (pt->exp_data->dependencies, depitem);
 		        if (configuration.verbose )
 		     g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "        ADDING DEPENDENCY: %s %s", 
		     depitem->type ,depitem->id);

		}
	      mgd_release (res);
	    }
	  cycle++;
	}
    }
}



static void
rpd_write_object_links_foreach (gpointer key, gpointer value, gpointer user_data)
{
   RListEntry *depitem;
   midgard_res *res;
   RSchemaField *val = value;
   gchar temptable[255], tempname[255], temptablename[255], tempidcol[255];
 
  struct
  {
    WExport *exp_data;
    int id;
  }
  *pt = user_data;
   res=NULL;
   
  if (!pt)
    return;
  if (configuration.verbose> 1) g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "  LINKS FOREACH %d on %d", val->type,pt->id);
  if (val->type == R_RLINK_FIELD)
    {
      (void) sscanf (key, "%s %s %s %s", temptable, tempname,
    		     temptablename, tempidcol);
    }
  else
    {
      (void) sscanf (key, "%s %s", temptable, tempname);
    }
   if (val->type == R_RLINK_FIELD)
     {
   	if (configuration.verbose> 1)
   	  g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "	  dependency look: %d %s %s %s %d %s %s", 
   	  val->type,   val->link,   temptable, tempname, pt->id, temptablename, tempidcol);					       
   	 
       /* NEEDS TO MAKE
   	  SELECT id FROM val->link WHERE temptablename = tempname AND tempidcol = id
   	*/
      if (!
       (configuration.nopasswords &&
       (!g_strcasecmp((gchar *) tempname, "person"))))  {
   	  res = mgd_sitegroup_select (pt->exp_data->mgd, "id",
	 			   val->link, "$s=$q AND $s=$d", NULL,
	 			   temptablename, tempname, temptable,
	 			   pt->id);
	 }			     
     }
   else
     {
      if (val->type == R_LINK_FIELD) {
        if (!
	      (configuration.nopasswords && 
	       (!g_strcasecmp((gchar *) val->link , "preference"))
	      ))  {
      
      
       res =
	 mgd_sitegroup_select (pt->exp_data->mgd, "id",
	 		       val->link, "$s=$d", NULL,
	 		       temptable, pt->id);
   
   	if (configuration.verbose > 1)
   	  g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "	  GETTING DEPS:%d  from %s where %s %d", 
   	    val->type,  val->link,   temptable,  pt->id);   
       }
       }      
   	 
     }
    if (res)
      {
    	while (mgd_fetch (res))
	  if(pt->exp_data->expandtree == TRUE) {
	    /* Add this link into the list of items requested for replication by dependency */
	    depitem = g_new (RListEntry, 1);
	    depitem->guid = FALSE;
	    depitem->type = g_strdup (val->link);
	    depitem->id = g_strdup (mgd_column (res, "id"));
	    g_list_append (pt->exp_data->dependencies, depitem);
	  	  if (configuration.verbose  )
	       g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "	 ADDING DEPENDENCY: %s %s", 
	       depitem->type ,depitem->id   );  					    
	      


	  }
    	mgd_release (res);
      }
 
}

static void
rpd_write_object_tree (gpointer value, gpointer user_data)
{

  WExport *exp_data = user_data;
  RListEntry *entry = value;
  midgard_res *res;
  midgard_pool *pool;
  GHashTable *link_hash;
  RSchemaType *object_type;
  int id, *tree, *cycle;
  gchar *table;

  if (!entry)
    return;

  if (entry->guid)
    {				/* Using GUID, gather id first */
      id = mgd_repligard_id (exp_data->mgd, entry->id);
      /* If it exists, replace GUID by table:id pair for common case */
      if (id)
	{
	  pool = mgd_alloc_pool ();
	  table = mgd_repligard_table (exp_data->mgd, pool, entry->id);
	  entry->guid = FALSE;
	  entry->type = g_strdup (table);
	  g_free (entry->id);
	  entry->id = g_strdup_printf ("%d", id);
	  mgd_free_pool (pool);
	}

    }

  /* Do not replicate site groups if configuration disables it */
  if(configuration.sitegroup == FALSE)
      if (!g_strcasecmp (entry->type, "sitegroup"))
    return;

  /* Now we are ready to work with subtree */
  object_type = repligard_get_object_type (exp_data->schema, entry->type);
  link_hash = (GHashTable *)
    g_hash_table_lookup (exp_data->schemacfg->links, entry->type);

  if ((object_type->type == R_TREE_OBJECT) && exp_data->expandtree)
    {
      tree = mgd_tree (exp_data->mgd, entry->type, "up", atoi (entry->id), 0, NULL);
      if (tree)
	{
	   cycle = tree;
	  while (cycle && (*cycle > 0))
	    {
	      res =
		mgd_sitegroup_record (exp_data->mgd, "*",
				      entry->type, *cycle);
	      if (res && mgd_fetch (res))
		{
		  repligard_write_object (object_type->
					  object, exp_data, entry->type, res);
		  mgd_release (res);
		}
	      cycle++;
	    }
	  /* Export objects which are linked with this one */
	  {
	    struct
	    {
	      WExport *exp_data;
	      int *tree;
	    }
	    pt;
	    pt.exp_data = exp_data;
	    pt.tree = tree;
	    g_hash_table_foreach (link_hash, rpd_write_links_foreach, &pt);
	  }
	  free (tree);
	}
    }
  else
    {
      WGuidCacheEntry  *entryinfo;
      gchar *guid;
     guid = g_strconcat ( entry->type , " ",  entry->id  , NULL);
     entryinfo = g_hash_table_lookup (exp_data->guid_hash, guid);
     
     if(entryinfo == NULL) {
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, 
	    "Repligard table is inconsistent!\n"
	    "Resource '%s' has no corresponding GUID and thus cannot be exported.\n"
	    "Please re-run repligard with '-m' option and export resources after that.\n"
	    "Repligard will continue this session in error checking mode...", guid);
	g_free(guid);
        return;
     }
     
     g_free(guid);
     if (entryinfo->replicated) return;
   
	
	 res =
	mgd_sitegroup_record (exp_data->mgd, "*", entry->type,
			      atoi (entry->id));
      if (res && mgd_fetch (res))
	{
	
	 repligard_write_object (object_type->object, exp_data,
				  entry->type, res);
	  mgd_release (res);
	  {
	  struct
	   {
	     WExport *exp_data;
	     int id;
	   }
	   pt;
	   pt.exp_data = exp_data;
	   pt.id = atoi (entry->id);
	   g_hash_table_foreach (link_hash, rpd_write_object_links_foreach, &pt);			  
	  } 
	}
    }
}

static void
rpd_update_guid_table (gpointer key, gpointer value, gpointer user_data)
{
  WExport *exp_data = user_data;
  WGuidCacheEntry *entry = value;

  if (entry->replicated)
    {

      (void) mgd_query (exp_data->mgd,
			"UPDATE repligard SET updated=NULL, changed=$s WHERE guid=$q",
			entry->changed, entry->guid);
    }
}

static void
rpd_export_deleted (gpointer key, gpointer value, gpointer user_data)
{
  WExport *exp_data = user_data;
  WGuidCacheEntry *entry = value;
  gchar temptable[255], tempname[255];

  if ((entry->deleted == TRUE) && (entry->replicated == FALSE))
    {
      (void) sscanf (key, "%s %s", temptable, tempname);

	gzprintf (exp_data->fd,
		    "<%s id=\"%s\" deleted=\"%s\"/>\n",
		    temptable, entry->guid, entry->changed);
	g_free (entry->updated);
	entry->updated = g_strdup (entry->changed);
	entry->replicated = TRUE;
    }
}

static void
rpd_remove_deleted (gpointer key, gpointer value, gpointer user_data)
{
  WExport *exp_data = user_data;
  WGuidCacheEntry *entry = value;

  if (entry->deleted == TRUE)
    {

      (void) mgd_query (exp_data->mgd,
			"UPDATE repligard set updated=NULL, changed=$s where guid=$q",
			entry->changed, entry->guid);
    }
}

static void
rpd_write_object_table (gpointer key, gpointer value, gpointer user_data)
{
  const gchar *table = key;
  WExport *exp_data = user_data;
  RSchemaType *object = value;
  gchar *guid;
  WObjectFD ofd;
  WObjectType otype;
  WGuidCacheEntry *entry;
  midgard_res *res = NULL;
  gchar  *usefrom="updated";
  gint export_result;

  if (configuration.fromdate)
    usefrom = configuration.fromdate;
  /* Do not replicate site groups if configuration disables it */
  if(configuration.sitegroup == FALSE)
      if (!g_strcasecmp (table, "sitegroup"))
	return;

  g_assert (exp_data != NULL);
  if (configuration.diff)
    {
      GList *items, *item;
#if HAVE_MIDGARD_SITEGROUPS      
	if (mgd_sitegroup(exp_data->mgd)) {
	    if (!g_strcasecmp(table, "sitegroup")) {
	       res = mgd_ungrouped_select(exp_data->mgd, "id,action", "repligard",
		   "changed > $p AND realm=$q AND sitegroup=$d", NULL, usefrom, table, 
		   mgd_sitegroup(exp_data->mgd));
	   
	    } else {
		res = mgd_ungrouped_select(exp_data->mgd, "id,action", "repligard",
 		   "changed > $p AND realm=$q AND sitegroup=$d", NULL, usefrom, table, 
		    mgd_sitegroup(exp_data->mgd));
 	    }		   
	} else {
#endif
	    res = mgd_ungrouped_select(exp_data->mgd, "id,action", "repligard",
		    "changed > $p AND realm=$q", NULL, usefrom, table);
#if HAVE_MIDGARD_SITEGROUPS      
	}					   
#endif

      if (!res)
	return;
      items = g_list_alloc ();
      while (mgd_fetch (res))
	{
	  /* Do not append deleted resources to list so they would not be fetched
	     because there is nothing to fetch actually
	  */
	  if(g_strcasecmp("delete", mgd_column(res, "action")) != 0) {
		g_list_append (items, GINT_TO_POINTER (mgd_sql2int (res, 0)));
	  } else {
		guid =
		    g_strconcat (table, " ", mgd_column (res, "id"), NULL);
		entry = g_hash_table_lookup (exp_data->guid_hash, guid);
		gzprintf (exp_data->fd,
			    "<%s id=\"%s\" deleted=\"%s\"/>\n",
			    table, entry->guid, entry->changed);
		g_free (entry->updated);
		entry->updated = g_strdup (entry->changed);
		entry->replicated = TRUE;
		g_free (guid);
	    
	  }
	}
      mgd_release (res);
      ofd.guid_hash = exp_data->guid_hash;
      ofd.exp_data = exp_data;
      item = g_list_first (items);
      while (item)
	{
	  if (GPOINTER_TO_INT (item->data) != 0)
	    {
	      res =
		mgd_sitegroup_select (exp_data->mgd, "*",
				      table, "id=$d", NULL,
				      GPOINTER_TO_INT (item->data));
	      if (res && mgd_fetch (res))
		{
		  ofd.res = res;
		  ofd.fd = exp_data->fd;
		  guid =
		    g_strconcat (table, " ", mgd_column (res, "id"), NULL);
		  entry = g_hash_table_lookup (exp_data->guid_hash, guid);
		  if (configuration.verbose)
#if HAVE_MIDGARD_SITEGROUPS	  
            	    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "exporting resource %s '%s' from SG%s", 
			table, guid, mgd_column(res, "sitegroup"));
#else
		    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "exporting resource %s '%s'",
			table, guid);
#endif
		  /* Export resource by default */
		  export_result = 0;
		  /* If export hook is set then run object through it */
		  if(configuration.export_hook != NULL) {
		    otype.res = res;
		    otype.otype = object->object;
		    otype.guid = entry->guid;
		    otype.ntype = table;
		    export_result = repligard_run_export_hook(&otype);
		  }
		  if(!export_result) {
		      gzprintf (exp_data->fd,
				"<%s id=\"%s\" changed=\"%s\">\n",
				table, entry->guid, entry->changed);
		      g_hash_table_foreach (object->
					    object,
					    rpd_write_object_field, &ofd);
		      gzprintf (exp_data->fd, "</%s>\n", table);
		  }
		  /* Set updated time to changed one to avoid cross-referenced exportation 
		     When export procedure will be done, this info will go back to DB 
		   */
		  g_free (entry->updated);
		  entry->updated = g_strdup (entry->changed);
		  entry->replicated = TRUE;
		  g_free (guid);
		  mgd_release (res);
		}
	    }
	  item = g_list_next (item);
	}
    }
  else
    {

#if HAVE_MIDGARD_SITEGROUPS
	if (mgd_sitegroup(exp_data->mgd)) {
	   if (!g_strcasecmp(table, "sitegroup")) {
                res = mgd_ungrouped_select(exp_data->mgd, "*", table,
		    "id=$d", NULL, mgd_sitegroup(exp_data->mgd));
	    } else {
                res = mgd_ungrouped_select(exp_data->mgd, "*", table,
		    "sitegroup=$d", NULL, mgd_sitegroup(exp_data->mgd));
	    }
	} else { 					    
#endif
	    res = mgd_sitegroup_select(exp_data->mgd, "*", table, NULL, NULL);
#if HAVE_MIDGARD_SITEGROUPS
        }
#endif

      /* Export deleted resources if any */
      g_hash_table_foreach (exp_data->guid_hash, rpd_export_deleted, &exp_data);
      
      if (!res)
	{
	  return;
	}

      ofd.guid_hash = exp_data->guid_hash;
      ofd.exp_data = exp_data;
      
      /* Export all other resources */
      while (mgd_fetch (res))
	{
	  ofd.res = res;
	  ofd.fd = exp_data->fd;
	  guid = g_strconcat (table, " ", mgd_column (res, "id"), NULL);
	  entry = g_hash_table_lookup (exp_data->guid_hash, guid);
	  if (entry)
	    {
	      /* Export resource by default */
	      export_result = 0;
	      /* If export hook is set then run object through it */
	      if(configuration.export_hook != NULL) {
		otype.res = res;
		otype.otype = object->object;
		otype.guid = entry->guid;
		otype.ntype = table;
		export_result = repligard_run_export_hook(&otype);
	      }
	      if(!export_result) {
		gzprintf (exp_data->fd,
			  "<%s id=\"%s\" changed=\"%s\">\n",
			  table, entry->guid, entry->changed);
		g_hash_table_foreach (object->object,
				      rpd_write_object_field, &ofd);
		gzprintf (exp_data->fd, "</%s>\n", table);
	      }
	      /* Set updated time to changed one to avoid cross-referenced exportation 
	         When export procedure will be done, this info will go back to DB 
	      */
	      g_free (entry->updated);
	      entry->updated = g_strdup (entry->changed);
	      entry->replicated = TRUE;
	    }
	  g_free (guid);
	}
      
      if (res)
	mgd_release (res);
    }
}

void
repligard_export_xml (midgard * mgd, RSchemaCfg * schemacfg,
		      GHashTable * guid_hash, gchar * filename)
{
  gzFile *fd;
  WExport exp_data;

  if (!filename)
    return;
  fd = gzopen (filename, "wb");
  if (!fd) {
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Cannot open file for write!");
    myexit(1);
  }
  exp_data.fd = fd;

  exp_data.mgd = mgd;
  exp_data.schemacfg = schemacfg;
  exp_data.schema = schemacfg->schema;
  exp_data.guid_hash = guid_hash;
  exp_data.dependencies = g_list_alloc ();
  exp_data.expandtree = TRUE;

  gzprintf (fd, "<?xml version=\"1.0\" encoding='%s'?>\n",
	    configuration.dbencoding);
  gzprintf (fd, "<Database xmlns=\"%s\">\n", R_SCHEMA_NS);

  if (configuration.all)
    {
      g_hash_table_foreach (schemacfg->schema, rpd_write_object_table,
			    &exp_data);
    }
  else
    {
      g_list_foreach (configuration.resource, rpd_write_object_tree,
		      &exp_data);
      if (configuration.verbose)
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "**************** WRITING DEPENDECIES ****************");
      /* Do not expand tree when replicate dependencies */
      exp_data.expandtree = FALSE;
      g_list_foreach (exp_data.dependencies, rpd_write_object_tree,
		      &exp_data);
    }
  gzprintf (fd, "</Database>\n");
  gzclose (fd);

  /* Free list of dependencies */
  g_list_foreach (exp_data.dependencies, rpd_remove_stringlist, NULL);
  g_list_free (exp_data.dependencies);

  /* Update entries which were replicated */
  g_hash_table_foreach (guid_hash, rpd_update_guid_table, &exp_data);
  /* Delete entries with action='delete' to clear repligard table */
  /* Note that we don't delete object themselves because they already deleted through Midgard's client
   */
  g_hash_table_foreach (guid_hash, rpd_remove_deleted, &exp_data);
}
