/* $Id: import.c,v 1.24 2002/12/08 03:25:52 armand Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    import: update logic for replication
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <errno.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <zlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include "midgard/midgard.h"
#include "schema.h"
#include "rpconfig.h"
#include "read.h"
#include "common.h"

typedef struct
{
	midgard *mgd;
	midgard_pool *pool;
	int sitegroup;
	gchar *fields, *values;
}
WQueryContainer;

#if HAVE_MIDGARD_SITEGROUPS
#define CREATE_REPLIGARD_GUID(mgd, table,id, guid, changed)\
	mgd_create_repligard(mgd, table, "guid,id,changed,updated,action,sitegroup",\
        "$q,$d,$q,$q,'update',$d",guid, id, changed, changed, mgd_sitegroup(mgd));

#define CREATE_REPLIGARD_GUID_FRESH(mgd, table,id, guid)\
	mgd_create_repligard(mgd, table, "guid,id,changed,action,sitegroup",\
        "$q,$d,19700101000000,'update',$d",guid, id, mgd_sitegroup(mgd));
#else
#define CREATE_REPLIGARD_GUID(mgd, table,id, guid, changed)\
	mgd_create_repligard(mgd, table, "guid,id,changed,updated,action",\
        "$q,$d,$q,$q,'update'",guid, id, changed, changed);
#define CREATE_REPLIGARD_GUID_FRESH(mgd, table,id, guid)\
	mgd_create_repligard(mgd, table, "guid,id,changed,action",\
        "$q,$d,19700101000000,'update'",guid, id);
#endif

static void rpd_make_insert(gpointer key, gpointer value, gpointer user_data)
{
	WQueryContainer *query = user_data;
	gchar *tmp;

#if HAVE_MIDGARD_SITEGROUPS
	/* Add the sitegroup column only if SG0 admin is logged into SG0 */
	if (
			g_strcasecmp((gchar *) key, "sitegroup") == 0
		&&
			!(mgd_isroot(query->mgd) && mgd_sitegroup(query->mgd) == 0)
	   ) return;
#endif	
	if (query->fields) {
		tmp =
		   mgd_format(query->mgd, query->pool, "$s, $s", query->fields,
			      key);
		mgd_free_from_pool(query->pool, query->fields);
	}
	else
		tmp = mgd_format(query->mgd, query->pool, "$s", key);
	query->fields = tmp;
	if (query->values) {
		tmp =
		   mgd_format(query->mgd, query->pool, "$s, $q", query->values,
			      value);
		mgd_free_from_pool(query->pool, query->values);
	}
	else
		tmp = mgd_format(query->mgd, query->pool, "$q", value);
	query->values = tmp;
        if (configuration.verbose > 1)
     	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
	   	 "Dump: %s | %s", (gchar*)key, (gchar*)value);
}

static void rpd_make_update(gpointer key, gpointer value, gpointer user_data)
{
	WQueryContainer *query = user_data;
	gchar *tmp;
	if (query->values) {
		tmp =
		   mgd_format(query->mgd, query->pool, "$s, $s=$q",
			      query->values, key, value);
		mgd_free_from_pool(query->pool, query->values);
	}
	else
		tmp = mgd_format(query->mgd, query->pool, "$s=$q", key, value);
	query->values = tmp;
}

static void
rpd_make_empty_insert(gpointer key, gpointer value, gpointer user_data)
{
	WQueryContainer *query = user_data;
	gchar *tmp;
	
#if HAVE_MIDGARD_SITEGROUPS
	/* Add the sitegroup column only if SG0 admin is logged into SG0 */
	if (
			g_strcasecmp((gchar *) key, "sitegroup") == 0
		&&
			!(mgd_isroot(query->mgd) && mgd_sitegroup(query->mgd) == 0)
	   ) return;
#endif	
	if (query->fields) {
		tmp =
		   mgd_format(query->mgd, query->pool, "$s, $s", query->fields,
			      key);
		mgd_free_from_pool(query->pool, query->fields);
	}
	else
		tmp = mgd_format(query->mgd, query->pool, "$s", key);
	query->fields = tmp;
	if (query->values) {
		tmp =
		   mgd_format(query->mgd, query->pool, "$s, ''", query->values);
		mgd_free_from_pool(query->pool, query->values);
	}
	else
		tmp = mgd_format(query->mgd, query->pool, "''");
	query->values = tmp;
}


/* Creates an empty object. Returns id of fake object or 0 for error
*/
int
repligard_create_fake_resource(WImport * imp_data,
			       gchar * object_name, gchar * guid)
{
	WQueryContainer query;
	WGuidCacheEntry *entry;
	int id;

	query.mgd = imp_data->mgd;
	query.pool = mgd_alloc_pool();
	query.fields = NULL;
	query.values = NULL;

	g_hash_table_foreach(repligard_get_object
			     (imp_data->schema, object_name),
			     rpd_make_empty_insert, &query);
	id =
	   mgd_create(imp_data->mgd, object_name, query.fields, "$s",
		      query.values);
	mgd_free_from_pool(query.pool, query.fields);
	mgd_free_from_pool(query.pool, query.values);
	/* If object created, we'll add it to guid_hash to be able to use it later on import phase */
	if (id) {
		CREATE_REPLIGARD_GUID_FRESH(imp_data->mgd, object_name, id,
					    guid);
		mgd_update_repligard(imp_data->mgd, object_name, id,
				     "updated=19700101000000");
		entry = g_new(WGuidCacheEntry, 1);
		entry->guid = g_strdup(guid);
		entry->changed =
		   g_strdup(mgd_repligard_changed_guid
			    (imp_data->mgd, query.pool, guid));
		/* Note that we don't delete allocated update because mgd_free_pool() will clear it anyway */
		entry->updated =
		   g_strdup(mgd_repligard_updated_guid
			    (imp_data->mgd, query.pool, guid));
		entry->replicated = FALSE;
		g_hash_table_insert(imp_data->guid_hash,
				    g_strdup_printf("%s %d", object_name, id),
				    entry);
		mgd_free_pool(query.pool);
		return id;
	}
	mgd_free_pool(query.pool);
	return 0;
}

void
repligard_import_resource(WImport * imp_data,
			  GHashTable * object, const gchar * object_name,
			  const gchar * guid, const gchar * changed)
{
	int id = mgd_repligard_id(imp_data->mgd, guid);
	int result = 0;
	gchar *locid;
	WGuidCacheEntry *entry;
	WQueryContainer query;


	query.mgd = imp_data->mgd;
	query.pool = mgd_alloc_pool();
	query.fields = NULL;
	query.values = NULL;
        if (configuration.verbose > 1)
    	         	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			   	 "importing %s %s", object_name,guid );

	if (id) {
	    /* If repligard record exists but object itself does not,
		force object creation and delete old repligard record
	    */
            if (!mgd_exists_id(imp_data->mgd, object_name , "id=$d", id)) {
		mgd_delete_repligard_guid(imp_data->mgd, guid);
	    	id = 0;
	    }
	}
	if (!id) {
		/* Resource does not exist in DB. Reason: */
		/* 1. This is new resource, we have to create it */
		g_hash_table_foreach(object, rpd_make_insert, &query);
		id =
		   mgd_create(imp_data->mgd, object_name, query.fields, "$s",
			      query.values);
	        if (configuration.verbose > 1)
         	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
		   	 "Object %s %s creation:\n / %s /\n %s", 
			    object_name, guid,
			    query.fields, query.values );
		if (id) {
			CREATE_REPLIGARD_GUID(imp_data->mgd, object_name, id,
					      guid, changed);
			mgd_update_repligard(imp_data->mgd, object_name, id,
					     "updated=NULL");
			locid = g_strdup_printf("%s %d", object_name, id);
			entry = g_new(WGuidCacheEntry, 1);
			entry->guid = g_strdup(guid);
			entry->changed = g_strdup(changed);
			/* Note that we don't delete allocated update because mgd_free_pool() will clear it anyway */
			entry->updated =
			   g_strdup(mgd_repligard_updated_guid
				    (imp_data->mgd, query.pool, guid));
			entry->replicated = FALSE;
			g_hash_table_insert(imp_data->guid_hash, locid, entry);
			mgd_free_pool(query.pool);
			query.pool = mgd_alloc_pool();
		        if (configuration.verbose > 1)
    	         	  	 g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			   	 "Object %s %s created successfuly", object_name,guid );

		}
	}
	else {
		/* Resource already in DB, check update and change times to solve situation */ 
		locid = g_strdup_printf("%s %d", object_name, id);
		entry = g_hash_table_lookup(imp_data->guid_hash, locid);
		if (!entry) {
		    repligard_build_guid_hash_add( imp_data->mgd, imp_data->guid_hash, guid );
		    entry = g_hash_table_lookup(imp_data->guid_hash, locid);
		}
		if (!entry) {
			/* Repligard record is corrupted (absent), construct it again */
			entry = g_new(WGuidCacheEntry, 1);
			entry->guid = g_strdup(guid);
			entry->changed = g_strdup(changed);
			entry->updated = g_strdup(changed);
			entry->replicated = FALSE;
			g_hash_table_insert(imp_data->guid_hash, locid, entry);
			mgd_free_pool(query.pool);
			query.pool = mgd_alloc_pool();
			CREATE_REPLIGARD_GUID(imp_data->mgd, object_name, id,
					      guid, changed);
			result = 1;
		}
		/* if updated < changed we have to update record */
		if (!entry->updated)
			g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING,
			      "entry->updated is NULL, updated/changed comparision might produce segfault");
		if ((configuration.diff == FALSE) ||
		    (g_strcasecmp(entry->updated, changed) < 0) ||
		    (result == 1)) {
			g_hash_table_foreach(object, rpd_make_update, &query);
			result =
			   mgd_update(imp_data->mgd, object_name, id, "$s",
				      query.values);
			mgd_free_from_pool(query.pool, query.values);
			if (result) {
				UPDATE_REPLIGARD_TIME(imp_data->mgd,
						      object_name, id, changed);
				mgd_update_repligard(imp_data->mgd, object_name,
						     id, "updated=NULL");
				/* Update guid_hash to contain most current timestamps */
				g_free(entry->changed);
				g_free(entry->updated);
				entry->changed =
				   g_strdup(mgd_repligard_changed_guid
					    (imp_data->mgd, query.pool, guid));
				entry->updated =
				   g_strdup(mgd_repligard_updated_guid
					    (imp_data->mgd, query.pool, guid));
				entry->replicated = TRUE;
		    		if (configuration.verbose > 1)
    	         	  	     g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
			   		"Object %s %s updated successfuly", object_name,guid );
			}
		}
		g_free(locid);
	}
	mgd_free_pool(query.pool);
}

static void
rpd_unlink_blob_foreach(gpointer key, gpointer value, gpointer user_data)
{
	RSchemaField *field = value;
	midgard_res *res = user_data;
	if (field) {
		if (field->type == R_BLOB_FIELD) {
			gchar *fname =
			   g_strdup_printf("%s/%s", configuration.blobdir,
					   mgd_column(res, key));
			if (fname)
				if (unlink(fname) != 0) {
					g_log(G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
					      "Error unlinking BLOB file '%s' for deleted resource. Reason: %s",
					      mgd_column(res, key),
					      strerror(errno));
					myexit(1);
				}
		}
	}
}

void
repligard_delete_resource(WImport * imp_data,
			  GHashTable * object, const gchar * object_name,
			  const gchar * guid, const gchar * deleted)
{
	/*
	   1. GHashTable * object actually empty because deleted resources
	   contains no fields.
	   2. const gchar * object_name is name of table in which resource should
	   be deleted.
	 */

	int id = mgd_repligard_id(imp_data->mgd, guid);

	if (id) {		/* Resource exists, we should delete it */
		RSchemaType *otype =
		   repligard_get_object_type(imp_data->schema, object_name);
		if (otype->type == R_FILE_OBJECT) {	/* BLOB object, external file should be unlinked */
			midgard_res *res =
			   mgd_sitegroup_record(imp_data->mgd, "*", object_name,
						id);
			if (res && mgd_fetch(res)) {
				g_hash_table_foreach(otype->object,
						     rpd_unlink_blob_foreach,
						     res);
				mgd_release(res);
			}
		}
		/*
		   1. Due iterative nature of repligard we can safely keep orphaned
		   entries non-touched. All resources deleted at source DB, Repligard will
		   mark as 'deleted' in dump file. Thus, they will be deleted too.
		   2. Even record_extensions and BLOBs linked to resource could be ignored
		   because they are resources too and will be deleted as usual
		   at some iteration.
		   3. So, all what we need is to know how to delete atomic resource
		 */

		mgd_delete(imp_data->mgd, object_name, id);
		mgd_delete_repligard(imp_data->mgd, object_name, id);
    		if (configuration.verbose > 1)
     	  	     g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
		     "Object %s %s deleted successfuly", object_name,guid );
	}
}
