/* $Id: read.c,v 1.20 2002/11/08 15:33:06 daniel Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *      read: reads Repligard file and calls update logic
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif

#define G_LOG_DOMAIN "REPLIGARD/READ"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <zlib.h>
#ifndef WIN32
#include <unistd.h>
#endif
#ifdef WIN32
#include <winsock2.h>
#endif
#include <mysql.h>
#include "midgard/midgard.h"
#include "midgard/internal.h"
#include "midgard/pageresolve.h"
#include "read.h"
#include "rpconfig.h"

#ifndef S_ISREG
#define S_ISREG(mode)	(((mode)&S_IFMT) == S_IFREG)
#endif

#ifndef WIN32
extern uid_t realuserid, suiduserid;
#endif

typedef enum
{
  R_PARSE_NONE,
  R_PARSE_DB,
  R_PARSE_OBJECT,
  R_PARSE_FIELD,
  R_PARSE_ERROR
}
RFileParserStatus;

typedef struct
{
  /* Hash table with Repligard schema 
     Each entry is a hash table for object, where object_type
     is the hash table for current object
   */
  GHashTable *schema, *object_type, *guid_hash;
  RSchemaCfg *schemacfg;
  midgard *mgd;
  XML_Parser parser;

  /* ICONV descriptor for UTF-8 -> DB encoding conversion */
  iconv_t cd;

  RFileParserStatus status;

  /* Object being parsed. We store object as hash table where
     each record corresponds one object's property.
     This is simple 'string -> string' correspondance
   */
  GHashTable *object;		/* Current object */

  /* Current object field: */
  RSchemaField *field;
  gchar *blob_locid;
  gchar *name;
  GByteArray *value;

  /* GUID of the current object, time stamp of its change, object name,
     and [optional] time stamp of delete time 
   */
  const gchar *guid, *changed, *object_name, *deleted;
  /* TRUE if object update must be delayed before update of other objects */
  gboolean delayed;
  /* GUIDs of objects that delayed this object 
     This table contains RDelayEntry objects hashed by GUIDs
   */
  GHashTable *objects;
}
RFile;

static void
rpd_read_elt_start (void *userdata, const char *eltname, const char **atts)
{
  RFile *data = userdata;
  gint nslen;

  nslen = strlen (R_SCHEMA_NS) + 1;
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s' required in line %d!", R_SCHEMA_NS,
	     XML_GetCurrentLineNumber (data->parser));
      myexit (1);
    }
  switch (data->status)
    {
    case R_PARSE_NONE:
      if (g_strcasecmp (eltname + nslen, R_TAG_READ_DB) == 0)
	{
	  data->status = R_PARSE_DB;
	}
      else
	data->status = R_PARSE_ERROR;
      break;
    case R_PARSE_DB:
      /* Check whether given object defined in our schema */
      data->object_name = g_strdup (eltname + nslen);
      data->object_type =
	repligard_get_object (data->schema, eltname + nslen);
      if (data->object_type != NULL)
	{
	  data->guid =
	    g_strdup (repligard_get_attribute (R_ATT_READ_ID, atts));
	  data->changed =
	    g_strdup (repligard_get_attribute (R_ATT_READ_CHANGED, atts));
	  data->deleted =
	    g_strdup (repligard_get_attribute (R_ATT_READ_DELETED, atts));
	  data->blob_locid = NULL;
	  if (configuration.verbose > 1)
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
		   "start object %s, %s", data->object_name, data->guid);
	  /* If GUID or CHANGED and DELETED underfined, trigger error */
	  if ((data->guid == NULL)
	      || ((data->changed == NULL) && (data->deleted == NULL)))
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Both 'id' and 'changed/deleted' attributes are required in object '%s' in line %d!",
		     eltname + nslen,
		     XML_GetCurrentLineNumber (data->parser));
	      myexit (1);
	    }
	  data->object = g_hash_table_new (rpd_honeyman_hash, g_str_equal);
	  g_assert (data->object != NULL);
	  /* OK, now we ready to parse object fields */
	  data->status = R_PARSE_OBJECT;
	}
      else
	{
	  /* Object of this type is undefined in schema, trigger error */
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Object '%s' in line %d is unknown to Repligard!",
		 eltname + nslen, XML_GetCurrentLineNumber (data->parser));
	  myexit (1);
	}
      break;
    case R_PARSE_OBJECT:
      /* Each object consists from series of object fields */
      /* Check whether given object field defined in our schema */
      data->field = repligard_get_field (data->object_type, eltname + nslen);
      if (data->field != NULL)
	{
	  data->name = g_strdup (eltname + nslen);
	  g_assert (data->name != NULL);
	  data->value = g_byte_array_new ();
	  g_assert (data->value != NULL);
	  data->status = R_PARSE_FIELD;
	  if (data->field->type == R_BLOB_FIELD)
	    {
	      data->blob_locid =
		g_strdup (repligard_get_attribute (R_ATT_BLOB_LOCID, atts));
	    }
	}
      else
	{
	  /* Object field of this type is undefined in schema, trigger error */
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Object field '%s' in line %d is unknown to Repligard!",
		 eltname + nslen, XML_GetCurrentLineNumber (data->parser));
	  myexit (1);
	}
      break;
    case R_PARSE_FIELD:
      /* Field processing is done at element's finish */
      break;
    default:
      data->status = R_PARSE_ERROR;
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Tag '%s' in line %d is unknown to Repligard!",
	     eltname + nslen, XML_GetCurrentLineNumber (data->parser));
      myexit (1);
    }
  return;
}

static gboolean
rpd_remove_string_foreach (gpointer key, gpointer value, gpointer user_data)
{
  if (key)
    g_free (key);
  if (value)
    g_free (value);
  return TRUE;
}

static void
rpd_read_elt_end (void *userdata, const char *eltname)
{
  RFile *data = userdata;
  gint nslen, len, id;
  gchar *temp;
  WImport imp_data;
  const char *mysql_err;

  nslen = strlen (R_SCHEMA_NS) + 1;
  /* Check namespace first. If it is not our case, bail out */
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s'  required in line %d!", R_SCHEMA_NS,
	     XML_GetCurrentLineNumber (data->parser));
      myexit (1);
    }
  switch (data->status)
    {
      /* Finishing database processing */
    case R_PARSE_DB:
      if (g_strcasecmp (eltname + nslen, R_TAG_READ_DB) == 0)
	{
	  data->status = R_PARSE_NONE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'Database' expected!");
	  myexit (1);
	}
      break;
      /* Finishing object processing */
    case R_PARSE_OBJECT:
      if (configuration.verbose > 1)
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
	       "got object %s", data->object_name);
      if (g_strcasecmp (eltname + nslen, data->object_name) == 0)
	{
	  imp_data.mgd = data->mgd;
	  imp_data.guid_hash = data->guid_hash;
	  imp_data.schema = data->schema;
	  imp_data.schemacfg = data->schemacfg;
	  if (data->deleted != NULL)
	    /*
	       Delete resource. We do not delete all dependant resources here
	       simply because this is handled in different way: Repligard already
	       put 'delete' commands for all dependancies during replication,
	       so we'll get this commands sooner or later while reading
	       this XML file. Because all tables are locked during replication,
	       deletion will be atomic for other processes and will complete
	       before Repligard's completion.
	     */
	    repligard_delete_resource (&imp_data,
				       data->object,
				       data->object_name,
				       data->guid, data->deleted);
	  else
	    {
	      /*
	         This is ordinary resource, just try to import it as is:
	         - if it exists, try to update it;
	         - if it doesn't exist, try to create it
	       */
	      repligard_import_resource (&imp_data,
					 data->object,
					 data->object_name,
					 data->guid, data->changed);

	    }
	  g_free ((void *) data->object_name);
	  data->object_name = NULL;
	  g_free ((void *) data->guid);
	  data->guid = NULL;
	  g_free ((void *) data->changed);
	  data->changed = NULL;
	  g_free ((void *) data->deleted);
	  data->deleted = NULL;
	  g_free ((void *) data->blob_locid);
	  data->blob_locid = NULL;
	  g_hash_table_foreach_remove (data->object,
				       rpd_remove_string_foreach, NULL);
	  g_hash_table_destroy (data->object);
	  data->object = NULL;
	  data->status = R_PARSE_DB;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for '%s' in line %d expected!",
		 data->object_name, XML_GetCurrentLineNumber (data->parser));
	  myexit (1);
	}
      break;
    case R_PARSE_FIELD:
      /* BLOB is special type of field which is BASE64-encoded, thus no iconv conversion is
         needed. We have to write it to file with name data->blob_locid relative
         to BlobDir (unless data->blob_locid already there and doesn't belong to
         this resource's GUID).
       */
      if (data->field->type == R_BLOB_FIELD)
	{
	  gchar *fname, *changed;
	  midgard_pool *pool;
	  struct stat filestat;
	  int result;

	  pool = mgd_alloc_pool ();
	  changed = mgd_repligard_updated_guid (data->mgd, pool, data->guid);

	  if ((changed == NULL)
	      || (g_strcasecmp (data->changed, changed) > 0))
	    {

	      fname =
		g_strdup_printf ("%s/%s",
				 configuration.blobdir, data->blob_locid);
#ifdef WIN32
	      result = stat (fname, &filestat);
#else
		  if (realuserid != 0) setuid(suiduserid);
	      result = stat (fname, &filestat);
	      if (realuserid != 0) setuid(realuserid);
#endif
	      if (result != -1)
		{
		  if (!S_ISREG (filestat.st_mode))
		    {
		      /* We can't write file, replication failed. */
		      g_log (G_LOG_DOMAIN,
			     G_LOG_LEVEL_CRITICAL,
			     "Cannot write target BLOB container '%s' because "
			     "it is not a regular file in '%s'.",
			     data->blob_locid, configuration.blobdir);
		      myexit (1);
		    }
		  else
		    {
		      /* We should select another name if this file doesn't belong to current GUID */
		      midgard_res *res;
		      int blobid = mgd_repligard_id (data->mgd,
						     data->guid);
		      if (blobid)
			{
			  res =
			    mgd_sitegroup_record
			    (data->mgd, "*", data->object_name, blobid);
			  if (res && mgd_fetch (res))
			    {
			      if (g_strcasecmp
				  (data->
				   blob_locid,
				   mgd_column (res, data->name)) != 0)
				{
				  /* A-ha, something changed: local BLOB location doesn't the same
				     as remote one. We ignore it and use local value.
				   */
				  g_free (data->blob_locid);
				  data->
				    blob_locid
				    = g_strdup (mgd_column (res, data->name));
				}
			      mgd_release (res);
			    }
			}
		      else
			{
			  /* Resource doesn't exist, but we'll prepare correct BLOB file for later
			     phase when Repligard will create BLOB resource itself.
			   */
			  gboolean l_ok = FALSE;
			  int idx = 0;
			  midgard_res *res;
			  /* Resource doesn't exist but BLOB file does. We have two ways to go:
			     1. BLOB file may be result of ghost operation. This is true, if 
			     blobs table contains no record with location field equal to data->blob_locid.
			     It might be result of manual deletion of resource (corresponding BLOB file
			     hasn't deleted) or BlobDir is shared between several databases which is
			     violation.

			     In this case we'll simply override this BLOB file because BlobDir
			     _must_ be different for different databases.

			     2. BLOB file has corresponding record in blobs table. We should select
			     different name.
			   */
			  /* Searching possible candidates. Note usage of ungrouped select operation
			     due need to look acros SiteGroups
			   */
			  res =
			    mgd_ungrouped_select
			    (data->mgd, "id",
			     data->object_name,
			     "$s = $q", NULL, data->name, data->blob_locid);
			  if (res && mgd_fetch (res))
			    {
			      /* We don't need all results, one hit will be enough to
			         initiate new name selection algorithm
			       */
			      mgd_release (res);
			      /* Name selection algorithm tries to find empty slot
			         named 'data->guid' in BlobDir's cache. It is pretty
			         random because combination of two directories and
			         GUID unlikely to be non-unique.

			         Note that algorithm is tightly connected with GUID's
			         nature. We assume that GUID contains only [0-9a-f]
			         characters, as it is with MD5 sums.
			       */
			      while ((l_ok !=
				      TRUE)
				     && (idx <
					 (MIDGARD_REPLIGARD_TAG_LENGTH
					  >> 1) - 2))
				{
				  g_free (fname);
				  fname =
				    g_strdup_printf
				    ("%s/%c/%c/%s",
				     configuration.blobdir,
				     toupper
				     (data->guid
				      [0
				       +
				       idx]),
				     toupper
				     (data->guid[1 + idx]), data->guid);
				  if (stat (fname, &filestat) == -1)
				    {
				      l_ok = TRUE;
				      break;
				    }
				  idx++;
				}
			      if (l_ok == TRUE)
				{	/* Name selected, copy it to data->blob_locid */
				  g_free (data->blob_locid);
				  data->
				    blob_locid
				    =
				    g_strdup
				    (fname
				     + strlen (configuration.blobdir) + 1);
				  g_free (fname);
				}
			      else
				{
				  g_log
				    (G_LOG_DOMAIN,
				     G_LOG_LEVEL_CRITICAL,
				     "Can't select new name for BLOB '%s'. Previous one"
				     " '%s' already occupied by other BLOB.",
				     data->guid, data->blob_locid);
				  myexit (1);
				}
			    }
			}
		    }
		}
	      /* OK, now we ready to decode BLOB into BlobDir/data->blob_locid */
	      repligard_decode_blob (data->value, data->blob_locid);
	    }
	  g_byte_array_free (data->value, TRUE);
	  temp = g_strdup (data->blob_locid);
	}
      else
	{
	  temp =
	    midgard_iconv (data->cd, data->value->data,
			   data->value->len, &len);
	  g_byte_array_free (data->value, TRUE);
	  switch (data->field->type)
	    {
	      /* Convert links to proper values */
	    case R_LINK_FIELD:
	    case R_LINK_FIELD_NODEPS:
	      {
		if (temp)
		  {
		    id = mgd_repligard_id (data->mgd, temp);
		    /* If resource does not exist, put it into dependency queue and delay object creation */
		    if (!id)
		      {
			imp_data.mgd = data->mgd;
			imp_data.guid_hash = data->guid_hash;
			imp_data.schema = data->schema;
			id =
			  repligard_create_fake_resource
			  (&imp_data, data->field->link, temp);
		      }
		    if (!id)
		      {
               mysql_err = mysql_error(imp_data.mgd->mysql);
               if (mysql_err == NULL) { mysql_err = "Unknown reason"; }
			g_log (G_LOG_DOMAIN,
			       G_LOG_LEVEL_CRITICAL,
			       "Failed to allocate delayed resource %s as a link %s for %s (%s).",
			       temp, data->name, data->object_name, mysql_err);
			myexit (1);
		      }
		    g_free (temp);
		    temp = g_strdup_printf ("%d", id);
		  }
		else
		  {
		    /* This is link to Administrator or to it is top level resource */
		    temp = g_strdup ("0");
		  }
	      }
	      break;
	    case R_RLINK_FIELD:
	      {
		if (temp)
		  {
		    id = mgd_repligard_id (data->mgd, temp);
		    /* If resource does not exist, put it into dependency queue and delay object creation */
		    if (!id)
		      {
			gchar temptable[255], tempname[255];
			gchar *actualtable;
			imp_data.mgd = data->mgd;
			imp_data.guid_hash = data->guid_hash;
			imp_data.schema = data->schema;
			(void) sscanf (data->
				       field->
				       link, "%s %s", temptable, tempname);
			actualtable =
			  (gchar *)
			  g_hash_table_lookup (data->object, temptable);
			if (!actualtable)
			  {
			    g_log
			      (G_LOG_DOMAIN,
			       G_LOG_LEVEL_CRITICAL,
			       "Failed to find a target table for delayed resource %s. "
			       "The field '%s' should be before RLINK in Schema for type %s!",
			       temp, temptable, data->object_name);
			    myexit (1);
			  }
			id =
			  repligard_create_fake_resource
			  (&imp_data, actualtable, temp);
		      }
		    if (!id)
		      {
               mysql_err = mysql_error(imp_data.mgd->mysql);
               if (mysql_err == NULL) { mysql_err = "Unknown reason"; }
			g_log (G_LOG_DOMAIN,
			       G_LOG_LEVEL_CRITICAL,
			       "Failed to allocate delayed resource %s as a link %s for %s (%s).",
			       temp, data->name, data->object_name, mysql_err);
			myexit (1);
		      }
		    g_free (temp);
		    temp = g_strdup_printf ("%d", id);
		  }
	      }
	      break;

	    default:
	      break;

	    }
	}

      g_hash_table_insert (data->object, data->name, temp);
      /* We aren't destroing these variables, they will be removed later,
         when object will be removed from the memory (see R_PARSE_OBJECT case)
       */
      data->status = R_PARSE_OBJECT;
      break;
      /* We don't handle this situations, they are internal for us but recent compilers complain
         about incomplete enumeration in switches
       */
    case R_PARSE_NONE:
    case R_PARSE_ERROR:
      break;
    }
  return;
}

void
rpd_read_cdata_add (void *userdata, const char *data, int len)
{
  RFile *xdata = userdata;
  /* All information is collected only at R_PARSE_FIELD level */
  if ((xdata != NULL)
      && (xdata->status == R_PARSE_FIELD) && (len > 0)
      && (xdata->value != NULL))
    {
      g_byte_array_append (xdata->value, data, len);
    }
}

static void
rpd_comment_handler (void *userData, const XML_Char * data)
{
  /* Do nothing, it allows to use comments in Schema file */
}


void
repligard_import_xml (midgard * mgd, RSchemaCfg * schemacfg,
		      GHashTable * guid_hash, const gchar * filename)
{


  XML_Parser parser;
  RFile data;
  gzFile *fp;
  guchar buf[BUFSIZ];
  gint done;
  if (configuration.verbose > 1)
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "IMPORTING FILE %s", filename);
  if ((fp = gzopen (filename, "r")) == NULL)
    {

      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
	     "ERROR OPENING FILE %s", filename);
      return;
    }
  /* We use it for converting data from UTF-8 to DB encoding as specified in configuration file */
  data.cd = iconv_open (configuration.dbencoding, "UTF-8");
  if ((size_t) data.cd == -1)
    {
      return;
    }
  if (configuration.verbose > 1)
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "ABOUT TO PARSE %s", filename);

  /*[eeh] this is pretty blunt but it'll work. Whatever type of
   * resource we import, we touch the cache. We could restrict this to
   * when we impport pages, styles, or style/page elements since these
   * are the bits that participate in the cache generation, but I
   * don't know enough about repligard to do that safely.
   */
  mgd_cache_touch(mgd, 0);

  parser = XML_ParserCreateNS (NULL, '#');
  g_assert (parser != NULL);
  data.schema = schemacfg->schema;
  data.schemacfg = schemacfg;
  data.mgd = mgd;
  data.parser = parser;
  data.name = NULL;
  data.value = NULL;
  data.object = NULL;
  data.guid_hash = guid_hash;

  XML_SetUserData (parser, &data);
  XML_SetElementHandler (parser, rpd_read_elt_start, rpd_read_elt_end);
  XML_SetCharacterDataHandler (parser, rpd_read_cdata_add);
  XML_SetCommentHandler (parser, rpd_comment_handler);
  /* Use this to allow localized comments and data */
  XML_SetUnknownEncodingHandler (parser, midgard_unknown_xml_encoding, NULL);

  data.status = R_PARSE_NONE;

  do
    {
      size_t len = gzread (fp, buf, sizeof (buf));
      done = len < sizeof (buf);
      if (!XML_Parse (parser, buf, len, done))
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "%s in line %d",
		 XML_ErrorString (XML_GetErrorCode
				  (parser)),
		 XML_GetCurrentLineNumber (parser));
	  if(data.cd != (iconv_t) -1) iconv_close (data.cd);
	  myexit (1);
	}

    }
  while (!done);

  XML_ParserFree (parser);
  iconv_close (data.cd);
  gzclose (fp);
}
