/* $Id: rpconfig.c,v 1.17 2002/03/12 13:49:15 ab Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *  rpconfig: reads Repligard configuration file
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif

#define G_LOG_DOMAIN "REPLIGARD/CONFIG"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include "midgard/midgard.h"
#include "midgard/tablenames.h"
#include "rpconfig.h"
#include "exporthook.h"

typedef enum
{
  R_PARSE_NONE,
  R_PARSE_INSIDE,
  R_PARSE_INCLUDE, 
  R_PARSE_DATABASE,
  R_PARSE_LOGIN,
  R_PARSE_REPLICATE,
  R_PARSE_EXPORTHOOK,
  R_PARSE_RESOURCE,
  R_PARSE_ERROR
}
RFileParserStatus;

typedef struct
{
  XML_Parser parser;

  RFileParserStatus status;

  gchar *schema;
  /* Database information */
  gchar *dbname, *dbusername, *dbpassword, *dbencoding;

  /* Login information */
  gchar *username, *password;

  /* Nested config file */
  gchar *includefile;

  /* Root directory for BLOBs */
  gchar *blobdir;

  /* Replication information */
  gboolean all;
  /* 
     This list keeps information about resources which 
     should be replicated. It includes resource type, local id
     and hashed by GUID of resource.
     If there are no <resource> sections in the <replicate>, then
     this hash table will be NULL. If 'all' attribute is set
     for <replicate>, then all resources will be replicated.
   */
  GList *resource;

  /* Export type support */
  GByteArray *hook_script;
  gchar *hook_type;
  iconv_t cd;
}
RFile;

RConfiguration configuration;

/* Converts external names of objects to the internal table names */
static const gchar* repligard_convert_type(const gchar *type) {
    int i;
    for(i=1; i<MIDGARD_OBJECT_COUNT; i++) {
	if(g_strcasecmp(mgd_table_extname[i], type) == 0) {
	    return mgd_table_name[i];
	}
    }
    return NULL;
}

static void
rpd_readconfig_elt_start (void *userdata, const char *eltname,
			  const char **atts)
{
  RFile *data = userdata;
  gint nslen;
  const gchar *id, *type, *st;
  RListEntry *entry;

  nslen = strlen (R_SCHEMA_NS) + 1;
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s' required in line %d!", R_SCHEMA_NS,
	     XML_GetCurrentLineNumber (data->parser));
	myexit(1);
    }
  switch (data->status)
    {
    case R_PARSE_NONE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_REPLIGARD) == 0)
	{
	  data->status = R_PARSE_INSIDE;
	}
      else
	data->status = R_PARSE_ERROR;
      break;
    case R_PARSE_INSIDE:
      /* Expect one of allowed top level configuration options */
      /* Is it <database>? */
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_DATABASE) == 0)
	{
	  data->status = R_PARSE_DATABASE;
	  data->schema =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_SCHEMA, atts));
	  data->dbname =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_NAME, atts));
	  data->dbusername =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_USERNAME, atts));
	  data->dbpassword =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_PASSWORD, atts));
	  data->dbencoding =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_ENCODING, atts));
	  data->blobdir =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_BLOBDIR, atts));
	  if(data->dbencoding != NULL) {
	    data->cd = iconv_open(data->dbencoding, "UTF-8");
	    if((size_t) data->cd == -1) {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Failed to make conversion from UTF-8 to database encoding '%s' in line %d!", data->dbencoding,
		     XML_GetCurrentLineNumber (data->parser));
	      myexit(1);
	    }
	  } else {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		   "You must specify database encoding using attribute '%s' in line %d!", R_ATT_RCFG_ENCODING,
		   XML_GetCurrentLineNumber (data->parser));
	    myexit(1);
	  }
	  return;
	}
      /* Is it <login>? */
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_LOGIN) == 0)
	{
	  data->status = R_PARSE_LOGIN;
	  data->username =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_USERNAME, atts));
	  data->password =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_PASSWORD, atts));
	  return;
	}
      /* Is it <replicate>? */
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_REPLICATE) == 0)
	{
	  data->status = R_PARSE_REPLICATE;
	  data->all = FALSE;
	  st = repligard_get_attribute (R_ATT_RCFG_ALL, atts);
	  if ((st != NULL) && (g_strcasecmp (st, "yes") == 0))
	    data->all = TRUE;
	  return;
	}
      /* Is it <exporthook>? */
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_EXPORTHOOK) == 0)
	{
	  data->status = R_PARSE_EXPORTHOOK;
	  data->hook_type = g_strdup(repligard_get_attribute (R_ATT_RCFG_TYPE, atts));
	  if(data->hook_type == NULL) {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		   "Attribute '%s' is required in line %d!", R_ATT_RCFG_TYPE,
		   XML_GetCurrentLineNumber (data->parser));
	    myexit(1);
	  }
	  /* We must know database encoding at this moment to be able correctly
	     convert script content to it because during export phase export hook
	     will operate with data in that encoding.
	   */
	  if(data->dbencoding == NULL) {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		   "Export hook definition must be after database encoding definition!");
	    myexit(1);
	  }
	  data->hook_script = g_byte_array_new();
	  return;
	}
      /* Is it <include>? */	
     if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_INCLUDE) == 0)
	{
	  data->status = R_PARSE_INCLUDE;
	  data->includefile =
	    g_strdup (repligard_get_attribute (R_ATT_RCFG_NAME, atts));
	  if (data->includefile) {
	     repligard_read_config (data->includefile);
	  }
	 return;
	}
	
      /* Bail out any unexpected options */
      data->status = R_PARSE_ERROR;
      break;
    case R_PARSE_REPLICATE:
      /* Inside <replicate> we support only one option -- <tree> */
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_RESOURCE) == 0)
	{
	  data->status = R_PARSE_RESOURCE;
	  /* If all='yes' specified, do not accept any <tree> section */
	  if (data->all == TRUE)
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "<replicate all='yes'> already specified, <tree> section is invalid");
		myexit(1);
	    }
	  if (data->resource == NULL)
	    data->resource = g_list_alloc ();
	  /* Insert this <resource> into queue. Resource description consists of
	     either guid or type:id pair. In both cases <resource> means that
	     all resources under this one will be used for replication
	   */
	  entry = g_new (RListEntry, 1);
	  if (!entry)
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Not enough memory to store replication info!");
		myexit(1);
	    }
	  id = repligard_get_attribute (R_ATT_RCFG_GUID, atts);
	  if (id != NULL)
	    {			/* GUID specified */
	      entry->type = NULL;
	      entry->guid = TRUE;
	      entry->id = g_strdup (id);
	    }
	  else
	    {
	      /* Pair 'type' 'id' specified */
	      id = repligard_get_attribute (R_ATT_RCFG_ID, atts);
	      type = repligard_get_attribute (R_ATT_RCFG_TYPE, atts);
	      if ((id == NULL) || (type == NULL))
		{
		  data->status = R_PARSE_ERROR;
		  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
			 "Section <%s> requires either 'guid' or pair 'id' and 'type' attributes in line %d!",
			 R_TAG_RCFG_RESOURCE,
			 XML_GetCurrentLineNumber (data->parser));
			myexit(1);
		}
	      entry->type = g_strdup (repligard_convert_type(type));
	      if(entry->type == NULL) {
		  data->status = R_PARSE_ERROR;
		  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
			 "Object type <%s> is incorrect in line %d!",
			 type,
			 XML_GetCurrentLineNumber (data->parser));
			myexit(1);
	      }
	      entry->id = g_strdup (id);
	      entry->guid = FALSE;
	    }
	  g_list_append (data->resource, entry);
	  return;
	}
      else
	{
	  data->status = R_PARSE_ERROR;
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Tag '%s' in line %d is unknown to Repligard!",
		 eltname + nslen, XML_GetCurrentLineNumber (data->parser));
		myexit(1);
	}
      break;
      /* Throw through because we don't accept anything inside <tree> */
    case R_PARSE_LOGIN:
    case R_PARSE_DATABASE:
    case R_PARSE_RESOURCE:
    default:
      data->status = R_PARSE_ERROR;
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Tag '%s' in line %d is unknown to Repligard!",
	     eltname + nslen, XML_GetCurrentLineNumber (data->parser));
	myexit(1);
    }
  return;
}

static void
rpd_readconfig_elt_end (void *userdata, const char *eltname)
{
  RFile *data = userdata;
  gint nslen, len;
  gchar *temp;

  nslen = strlen (R_SCHEMA_NS) + 1;
  /* Check namespace first. If it is not our case, bail out */
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s'  required! (Line %d)",
	     R_SCHEMA_NS, XML_GetCurrentLineNumber (data->parser));
	myexit(1);
    }
  switch (data->status)
    {
      /* Finishing Repligard processing */
    case R_PARSE_INSIDE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_REPLIGARD) == 0)
	{
	  data->status = R_PARSE_NONE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'Repligard' expected!");
	myexit(1);
	}
      break;
      /* Finishing options processing */
    case R_PARSE_DATABASE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_DATABASE) == 0)
	{
	  data->status = R_PARSE_INSIDE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'database' expected!");
	myexit(1);
	}
      break;
    case R_PARSE_LOGIN:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_LOGIN) == 0)
	{
	  data->status = R_PARSE_INSIDE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'login' is expected!");
	myexit(1);
	}
      break;
    case R_PARSE_REPLICATE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_REPLICATE) == 0)
	{
	  data->status = R_PARSE_INSIDE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'replicate' is expected!");
	myexit(1);
	}
      break;
    case R_PARSE_EXPORTHOOK:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_EXPORTHOOK) == 0)
	{
	  data->status = R_PARSE_INSIDE;
	  temp = midgard_iconv(data->cd, data->hook_script->data, data->hook_script->len, &len);
	  g_byte_array_free(data->hook_script, TRUE);
	  repligard_add_export_type(data->hook_type, temp);
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'exporthook' is expected!");
	myexit(1);
	}
      break;
    case R_PARSE_RESOURCE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_RESOURCE) == 0)
	{
	  data->status = R_PARSE_REPLICATE;
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'resource' is expected!");
	myexit(1);
	}
      break;   
   case R_PARSE_INCLUDE:
      if (g_strcasecmp (eltname + nslen, R_TAG_RCFG_INCLUDE) == 0)
	{
	  data->status = R_PARSE_INSIDE;
          g_free(data->includefile);
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "Mismatched sequence: finishing tag for 'include' is expected!");
	myexit(1);
	}
      break;
    case R_PARSE_NONE:
    case R_PARSE_ERROR:
      break;
    }
  return;
}

void
rpd_read_cdata_hook (void *userdata, const char *data, int len)
{
  RFile *xdata = userdata;
  /* All information is collected only at R_PARSE_FIELD level */
  if ((xdata != NULL)
      && (xdata->status == R_PARSE_EXPORTHOOK) && (len > 0)
      && (xdata->hook_script != NULL))
    {
      g_byte_array_append (xdata->hook_script, data, len);
    }
}


static void
rpd_comment_handler (void *userData, const XML_Char * data)
{
  /* Do nothing, it allows to use comments in config file */
}

void
rpd_remove_stringlist (gpointer value, gpointer user_data)
{
  RListEntry *entry = value;
  if (entry != NULL)
    {
      /* This is just for test
         if(entry->guid) 
         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "GUID::%s",entry->id);
         else 
         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "%s::%s",entry->type, entry->id);
       */
      g_free (entry->type);
      g_free (entry->id);
      g_free (entry);
    }
}

/* Reads Repligard configuration file and merges it with previous one (if it exists).
   Does correct merge of replication trees and resource (de)allocation.
*/
void
repligard_read_config (const gchar * filename)
{

  XML_Parser parser;
  RFile data;
  gzFile *fp;
  guchar buf[BUFSIZ];
  gint done;
  if ((fp = gzopen (filename, "r")) == NULL)
    {
      return;
    }
  parser = XML_ParserCreateNS (NULL, '#');
  g_assert (parser != NULL);
  data.parser = parser;
  data.dbname = data.dbusername = data.dbpassword =
    data.username = data.password = NULL;
  data.dbencoding = NULL;
  data.includefile = NULL;
  data.resource = NULL;
  data.all = FALSE;
  data.schema = NULL;
  data.blobdir = NULL;
  data.hook_type = NULL;
  data.hook_script = NULL;
  data.cd = (iconv_t) -1;
 
  XML_SetUserData (parser, &data);
  XML_SetCommentHandler (parser, rpd_comment_handler);
  XML_SetElementHandler (parser,
			 rpd_readconfig_elt_start, rpd_readconfig_elt_end);
  /* Use this to allow localized comments */
  XML_SetUnknownEncodingHandler (parser, midgard_unknown_xml_encoding, NULL);
  XML_SetCharacterDataHandler(parser, rpd_read_cdata_hook);
  data.status = R_PARSE_NONE;
  do
    {
      size_t len = gzread (fp, buf, sizeof (buf));
      done = len < sizeof (buf);
      if (!XML_Parse (parser, buf, len, done))
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 "%s in line %d",
		 XML_ErrorString (XML_GetErrorCode (parser)),
		 XML_GetCurrentLineNumber (parser));
	if(data.cd != (iconv_t) -1) iconv_close (data.cd);
	myexit(1);
	}

    }
  while (!done);
  XML_ParserFree (parser);
  gzclose (fp);
  
  if(data.cd != (iconv_t) -1) iconv_close (data.cd);
  
  /* Update configuration. This allows nested configuration files, for example,
     to separate global config with DB setup and local configurations for
     different replicable areas. This complicates code a bit :-)
   */
#define UPDATE_CFG(var) \
  if(data.var != NULL) { \
    if(configuration.var != NULL) g_free(configuration.var); \
    configuration.var = data.var; \
  }
  UPDATE_CFG (schema);
  UPDATE_CFG (dbname);
  UPDATE_CFG (dbusername);
  UPDATE_CFG (dbpassword);
  UPDATE_CFG (dbencoding);
  UPDATE_CFG (blobdir);
  UPDATE_CFG (username);
  UPDATE_CFG (password);
  if (data.all == TRUE)
    {
      /* Oops, we have to delete previously allocated resources because this one overrides it 
         by full database replication.
       */
      if ((configuration.all == FALSE) && (configuration.resource != NULL))
	{
	  g_list_foreach (configuration.resource,
			  rpd_remove_stringlist, NULL);
	  configuration.resource = NULL;
	}
      configuration.all = TRUE;
    }
  else
    {
      if (data.resource != NULL)
	{
	  configuration.all = FALSE;
	  if (configuration.resource != NULL)
	    configuration.resource =
	      g_list_concat (configuration.resource, data.resource);
	  else
	    configuration.resource = data.resource;
	}
    }
}

void
repligard_init (void)
{
#define INIT_CFG(var) configuration.var = NULL;
  INIT_CFG (schema);
  INIT_CFG (dbname);
  INIT_CFG (dbusername);
  INIT_CFG (dbpassword);
  INIT_CFG (dbencoding);
  INIT_CFG (blobdir);
  INIT_CFG (username);
  INIT_CFG (password);
  INIT_CFG (resource);
  INIT_CFG (fromdate);
  INIT_CFG (export_hook);
  INIT_CFG (export_types);

  configuration.all = FALSE;
  configuration.verbose = 0;
  configuration.diff = TRUE;
  configuration.sitegroup = FALSE;
  configuration.nopasswords= FALSE;
}

void
repligard_done (void)
{
#define DONE_CFG(var) { \
    if(configuration.var != NULL) g_free(configuration.var); \
     configuration.var = NULL; \
  }
  DONE_CFG (schema);
  DONE_CFG (dbname);
  DONE_CFG (dbusername);
  DONE_CFG (dbpassword);
  DONE_CFG (dbencoding);
  DONE_CFG (blobdir);
  DONE_CFG (username);
  DONE_CFG (password);
  g_list_foreach (configuration.resource, rpd_remove_stringlist, NULL);
  g_list_free (configuration.resource);
  configuration.resource = NULL;
  configuration.all = FALSE;
  configuration.diff = TRUE;
  configuration.sitegroup = FALSE;
  configuration.export_hook = NULL;
  repligard_destroy_export_types();
}

