/* $Id: schema.c,v 1.19 2002/03/12 13:49:15 ab Exp $
 *
 * repligard: Replication utility for Midgard databases.
 * ---------:
 *    schema: utility functions for parsing Repligard config file
 *
 * Copyright (C) 2000 The Midgard Project Ry
 * Copyright (C) 2000 Alexander Bokovoy - Aurora S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif

#define G_LOG_DOMAIN "REPLIGARD/SCHEMA"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <zlib.h>
#include "midgard/midgard.h"
#include "schema.h"
#include "rpconfig.h"

typedef enum
{
  R_PARSE_NONE,
  R_PARSE_SCHEMA,
  R_PARSE_TYPE,
  R_PARSE_FIELD,
  R_PARSE_LINK,
  R_PARSE_ERROR
}
RSchemaParserStatus;

typedef struct
{
  /* Hash table 'types' contains hashes for all Repligard record's types 
     Each record type is a hash table itself and contains types of its components
     hashed by their names */
  GHashTable *types;
  XML_Parser parser;

  RSchemaParserStatus status, prevst;

  RSchemaType *type;		/* Current record type */
  RSchemaField *field;		/* Current record field */
  gchar *object_name;
}
RSchema;

/*
 * This is a simplified version of the pathalias hashing function.
 * Thanks to Steve Belovin and Peter Honeyman
 *
 * hash a string into a long int.  31 bit crc (from andrew appel).
 * the crc table is computed at run time by crcinit() -- we could
 * precompute, but it takes 1 clock tick on a 750.
 *
 * This fast table calculation works only if POLY is a prime polynomial
 * in the field of integers modulo 2.  Since the coefficients of a
 * 32-bit polynomial won't fit in a 32-bit word, the high-order bit is
 * implicit.  IT MUST ALSO BE THE CASE that the coefficients of orders
 * 31 down to 25 are zero.  Happily, we have candidates, from
 * E. J.  Watson, "Primitive Polynomials (Mod 2)", Math. Comp. 16 (1962):
 *	x^32 + x^7 + x^5 + x^3 + x^2 + x^1 + x^0
 *	x^31 + x^3 + x^0
 *
 * We reverse the bits to get:
 *	111101010000000000000000000000001 but drop the last 1
 *         f   5   0   0   0   0   0   0
 *	010010000000000000000000000000001 ditto, for 31-bit crc
 *	   4   8   0   0   0   0   0   0
 */

#define POLY 0x48000000L	/* 31-bit polynomial (avoids sign problems) */

static guint CrcTable[128];
static guint8 CrcInit = 0;

/*
 - crcinit - initialize tables for hash function
 */
static void
rpd_crc_init (void)
{
  int i, j;
  guint sum;

  for (i = 0; i < 128; ++i)
    {
      sum = 0L;
      for (j = 7 - 1; j >= 0; --j)
	if (i & (1 << j))
	  sum ^= POLY >> j;
      CrcTable[i] = sum;
    }
  CrcInit = 1;
}

/*
 - hash - Honeyman's nice hashing function
 */
guint
rpd_honeyman_hash (gconstpointer key)
{
  const gchar *name = (const gchar *) key;
  gint size;
  guint sum = 0;

  g_assert (name != NULL);
  g_assert (*name != 0);

  if (CrcInit != 0)
    rpd_crc_init ();

  size = strlen (name);

  while (size--)
    {
      sum = (sum >> 7) ^ CrcTable[(sum ^ (*name++)) & 0x7f];
    }

  return (sum);
}

const gchar *
repligard_get_attribute (const char *name, const char **atts)
{
  if (atts == NULL)
    return NULL;

  while (*atts)
    {
      if (g_strcasecmp (name, *atts) == 0)
	return *(atts + 1);
      atts += 2;
    }

  return NULL;
}

static struct
{
  const gchar *s;
  RSchemaFieldType type;
}
rpd_field_types[] =
{
  {"string", R_STRING_FIELD},
  {"text", R_TEXT_FIELD},
  {"integer", R_INTEGER_FIELD},
  {"password", R_PASSWORD_FIELD},
  {"link", R_LINK_FIELD},
  {"rlink", R_RLINK_FIELD},
  {"blob", R_BLOB_FIELD},
  {NULL, R_STRING_FIELD}
};

RSchemaFieldType
repligard_field_from_string (const gchar * str)
{
  gint i;

  for (i = 0; rpd_field_types[i].s; i++)
    if (g_strcasecmp (str, rpd_field_types[i].s) == 0)
      return rpd_field_types[i].type;

  return rpd_field_types[i].type;

}

const gchar *
repligard_string_from_field (RSchemaFieldType t)
{
  gint i;

  for (i = 0; rpd_field_types[i].type; i++)
    if (t == rpd_field_types[i].type)
      return rpd_field_types[i].s;

  return rpd_field_types[i].s;

}

void
rpd_schema_elt_start (void *userdata, const char *eltname, const char **atts)
{
  RSchema *schema = (RSchema *) userdata;
  const gchar *attname, *atttype, *attreverse;
  gint nslen;

  nslen = strlen (R_SCHEMA_NS) + 1;
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s' required in line %d!", R_SCHEMA_NS,
	     XML_GetCurrentLineNumber (schema->parser));
	myexit(1);
      return;
    }
  switch (schema->status)
    {
    case R_PARSE_NONE:
      if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_SCHEMA) == 0)
	{
	  schema->status = R_PARSE_SCHEMA;
	}
      else
	schema->status = R_PARSE_ERROR;
      break;
    case R_PARSE_SCHEMA:
      if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_TYPE) == 0)
	{
	  attname = repligard_get_attribute (R_ATT_SCHEMA_NAME, atts);
	  if ((attname == NULL) || (*attname == '\0'))
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Type name required for type definition!");
		myexit(1);
	    }
	  else
	    {
	      /* Add new record type and make it current */
	      schema->type = g_new (RSchemaType, 1);
	      schema->type->object =
		g_hash_table_new (rpd_honeyman_hash, g_str_equal);
	      g_assert (schema->type->object != NULL);
	      /* Assume flat objects by default */
	      schema->type->type = R_FLAT_OBJECT;
	      if (configuration.verbose > 1)
               g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "ADDING OBJ %s  ", 
			    		attname  );   	
 
	      g_hash_table_insert (schema->types,
				   g_strdup (attname), schema->type);
	      schema->object_name = g_strdup (attname);
	      schema->status = R_PARSE_TYPE;
	    }
	}
      break;
    case R_PARSE_TYPE:
      /* There are two field types: field and link */
      /* First: test against regular field */
      if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_FIELD) == 0)
	{
	  attname = repligard_get_attribute (R_ATT_SCHEMA_NAME, atts);
	  atttype = repligard_get_attribute (R_ATT_SCHEMA_TYPE, atts);
	  if ((attname == NULL) || (*attname == '\0')
	      || (atttype == NULL) || (*atttype == '\0'))
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Field name/type required for field definition!");
		myexit(1);
	    }
	  else
	    {
	      /* Add new record field and make it current */
	      schema->field = g_new (RSchemaField, 1);
	      g_assert (schema->field);
	      /* Fill in field type */
	      schema->field->type = repligard_field_from_string (atttype);
	      g_hash_table_insert (schema->type->object,
				   g_strdup (attname), schema->field);
	    }
	  /* Second: check is it link */
	}
      else if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_LINK) == 0)
	{
	  attname = repligard_get_attribute (R_ATT_SCHEMA_NAME, atts);
	  atttype = repligard_get_attribute (R_ATT_SCHEMA_LINK, atts);
	  attreverse = repligard_get_attribute (R_ATT_SCHEMA_REVERSE, atts);
	  if ((attname == NULL) || (*attname == '\0')
	      || (atttype == NULL) || (*atttype == '\0'))
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Field name/link required for field definition!");
		myexit(1);
	    }
	  else
	    {
	      /* Add new record field and make it current */
	      schema->field = g_new (RSchemaField, 1);
	      g_assert (schema->field);
	      /* Fill in field type */
	      schema->field->type = R_LINK_FIELD;
	      schema->field->link = g_strdup (atttype);

	      /* This is the link to itself, so the object is R_TREE_OBJECT */
	      if (g_strcasecmp (schema->object_name, atttype) == 0)
		{
		  schema->type->type = R_TREE_OBJECT;
		}
	      /* Igonre reverse dependencies for fields with 'reverse="no"' */
	      if ((attreverse != NULL) 
		  && (g_strcasecmp (R_ATT_SCHEMA_REVERSE_VAL, attreverse) == 0))
		{
		 
		  schema->field->type = R_LINK_FIELD_NODEPS;
		}
	      	if (configuration.verbose > 1)
             g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Storing Link:   %s %d", 
			     schema->field->link ,schema->field->type    );   
	      g_hash_table_insert (schema->type->object,
				   g_strdup (attname), schema->field);
	    }
	}
      else if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_RLINK) == 0)
	{
	  attname = repligard_get_attribute (R_ATT_SCHEMA_NAME, atts);
	  atttype = repligard_get_attribute (R_ATT_SCHEMA_LINK, atts);
	  if ((attname == NULL) || (*attname == '\0')
	      || (atttype == NULL) || (*atttype == '\0'))
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     "Field name/link required for field definition!");
		myexit(1);
	    }
	  else
	    {
	      /* Add new record field and make it current */
	      schema->field = g_new (RSchemaField, 1);
	      g_assert (schema->field);
	      /* Fill in field type */
	      schema->field->type = R_RLINK_FIELD;
	      schema->field->link = g_strdup (atttype);
	      /* This is the R_ACTIVE_OBJECT since it has R_RLINK_FIELD */
	      schema->type->type = R_ACTIVE_OBJECT;
	      if (configuration.verbose > 1)
               g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Storing RLINK:   %s %d ", 
			     schema->field->link ,schema->field->type    );   
	      g_hash_table_insert (schema->type->object,
				   g_strdup (attname), schema->field);
	    }
	}
      break;
    default:
      schema->status = R_PARSE_ERROR;
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Unexpected error in line %d!",
	     XML_GetCurrentLineNumber (schema->parser));
	myexit(1);

    }
}

void
rpd_schema_elt_end (void *userdata, const char *eltname)
{
  RSchema *schema = (RSchema *) userdata;
  gint nslen;

  nslen = strlen (R_SCHEMA_NS) + 1;
  if (strstr (eltname, R_SCHEMA_NS) != eltname)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     "Name space '%s' required in line %d!", R_SCHEMA_NS,
	     XML_GetCurrentLineNumber (schema->parser));
	myexit(1);
    }
  if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_TYPE) == 0)
    {
      g_free (schema->object_name);
      schema->status = R_PARSE_SCHEMA;
      return;
    }
  if ((g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_FIELD) == 0) ||
      (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_LINK) == 0) ||
      (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_RLINK) == 0))
    {
      schema->status = R_PARSE_TYPE;
      return;
    }
  if (g_strcasecmp (eltname + nslen, R_TAG_SCHEMA_SCHEMA) == 0)
    {
      schema->status = R_PARSE_NONE;
      return;
    }
  schema->status = R_PARSE_ERROR;
  return;
}

static void
rpd_comment_handler (void *userData, const XML_Char * data)
{
  /* Do nothing, it allows to use comments in Schema file */
}

/* Fill information about each dependency link in the manner the same to Schema */
/* Each object has a set of fields which are links to it by another objects */
/* These links are indexed by 'name type' string where 'name' is the name of the 
   link in the object 'type'. It allows multiple links from one object and
   multiple links with the same name from different objects
*/

static void
rpd_fill_onerlink_foreach (gpointer key, gpointer value, gpointer user_data)
{
  RSchemaField *field;
  GHashTable *onelink;
  RSchemaField *linkfield;
  struct
  {
    GHashTable *links;
    gchar *name;
    gchar *link;

    gchar *objname;
    RSchemaCfg *rschemacfg;

  }
   *pt = user_data;
   field = (RSchemaField*) value;
  /* now it needs to find out if the pt->name element has a RLINK! */
  onelink = (GHashTable *) g_hash_table_lookup (pt->links, key);
  if (!onelink)
    return;
  linkfield = g_new (RSchemaField, 1);
  linkfield->type = R_RLINK_FIELD;
  linkfield->link = g_strdup (pt->objname);
  if (configuration.verbose > 1)
           g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "reverse linking %s %s %s ", 
			    		pt->name, (gchar*) key , pt->link );   	
  g_hash_table_insert (onelink,
		       g_strdup_printf ("%s %s %s", pt->name, (gchar*) key, pt->link),
		       linkfield);
}

static void
rpd_fill_onelink_foreach (gpointer key, gpointer value, gpointer user_data)
{
  RSchemaField *field = value;
  struct
  {
    GHashTable *links;
    gchar *name;
    RSchemaCfg *rschemacfg;

  }
   *pt = user_data;
  RSchemaCfg *schemacfg = pt->rschemacfg;
  GHashTable *onelink;
  RSchemaField *linkfield;
  struct
  {
    GHashTable *links;
    gchar *name;
    gchar *link;
    gchar *objname;
    RSchemaCfg *rschemacfg;

  }
  pb;

  pb.links = pt->links;
  pb.rschemacfg = pt->rschemacfg;
  pb.name = key;
  pb.objname = pt->name;



  /* If field is the link */
  if (field && (field->type == R_RLINK_FIELD))
    {
      /*  if (configuration.verbose)
         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "RLINK LOOKUP in link %s", 
         field->link  );
       */

      pb.link = field->link;
      g_hash_table_foreach (schemacfg->schema, rpd_fill_onerlink_foreach,
			    &pb);
    }


  if (field && (field->type == R_LINK_FIELD))
    {
      /* Get target table and add a link to the processed one */
      /*   if (configuration.verbose)
         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "LINK LOOKUP in link %s", 
         field->link  );
       */
      onelink = (GHashTable *) g_hash_table_lookup (pt->links, field->link);
      /* If onlink is empty it probably means that link is defined incorrectly in the Schema file */
      if (!onelink)
	return;
      linkfield = g_new (RSchemaField, 1);
      linkfield->type = R_LINK_FIELD;
      linkfield->link = g_strdup (pt->name);
      /*
         if (configuration.verbose)
         g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "ADDING  in link %s %s", 
         key, pt->name     );
       */
      g_hash_table_insert (onelink,
			   g_strdup_printf ("%s %s", (gchar*) key, pt->name),
			   linkfield);
    }
}

static void
rpd_fill_links_foreach (gpointer key, gpointer value, gpointer user_data)
{

  RSchemaCfg *rschemacfg = user_data;
  RSchemaType *type = value;
  GHashTable *links = rschemacfg->links;
  struct
  {
    GHashTable *links;
    gchar *name;
    RSchemaCfg *rschemacfg;

  }
  pt;
  pt.links = links;
  pt.name = key;
  pt.rschemacfg = rschemacfg;

  if (type && type->object)
    {
      g_hash_table_foreach (type->object, rpd_fill_onelink_foreach, &pt);
    }
}

/* Initialize a links hash table by creating empty hashes for each objects */
static void
rpd_init_links_foreach (gpointer key, gpointer value, gpointer user_data)
{
  GHashTable *links = user_data;
  GHashTable *link = g_hash_table_new (rpd_honeyman_hash, g_str_equal);
  g_hash_table_insert (links, g_strdup (key), link);
}

/* Parses Repligard schema file and returns hash of hashes 
   describing its structure */
RSchemaCfg *
repligard_init_schema (const gchar * filename)
{

  XML_Parser parser;
  RSchema schema;
  RSchemaCfg *schemacfg;

  gzFile *fp;
  guchar buf[BUFSIZ];
  gint done;

  if ((fp = gzopen (filename, "r")) == NULL)
    {
      return NULL;
    }

  rpd_crc_init ();
  
  schema.types = g_hash_table_new (rpd_honeyman_hash, g_str_equal);
  g_assert (schema.types != NULL);

  parser = XML_ParserCreateNS (NULL, '#');
  g_assert (parser != NULL);
  schema.parser = parser;

  XML_SetUserData (parser, &schema);
  XML_SetElementHandler (parser, rpd_schema_elt_start, rpd_schema_elt_end);
  XML_SetCommentHandler (parser, rpd_comment_handler);
  XML_SetUnknownEncodingHandler (parser, midgard_unknown_xml_encoding, NULL);

  schema.status = R_PARSE_NONE;

  do
    {
      size_t len = gzread (fp, buf, sizeof (buf));
      done = len < sizeof (buf);

      if (!XML_Parse (parser, buf, len, done))
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "%s in line %d",
		 XML_ErrorString (XML_GetErrorCode (parser)),
		 XML_GetCurrentLineNumber (parser));
	myexit(1);
	}

    }
  while (!done);

  XML_ParserFree (parser);
  gzclose (fp);

  schemacfg = g_new (RSchemaCfg, 1);
  schemacfg->schema = schema.types;
  schemacfg->links = g_hash_table_new (rpd_honeyman_hash, g_str_equal);
  g_hash_table_foreach (schemacfg->schema, rpd_init_links_foreach,
			schemacfg->links);
  g_hash_table_foreach (schemacfg->schema, rpd_fill_links_foreach, schemacfg);

  return schemacfg;


}

static gboolean
rpd_remove_field_foreach (gpointer key, gpointer value, gpointer user_data)
{
  RSchemaField *field;
  field = (RSchemaField *) value;
  if ((field->type == R_LINK_FIELD) || (field->type == R_RLINK_FIELD))
    g_free (field->link);
  g_free (value);
  g_free (key);
  return TRUE;
}

static gboolean
rpd_remove_type_foreach (gpointer key, gpointer value, gpointer user_data)
{
  RSchemaType *type = (RSchemaType *) value;
  if (type)
    {
      if (type->object)
	{
	  g_hash_table_foreach_remove (type->object,
				       rpd_remove_field_foreach, NULL);
	  g_hash_table_destroy (type->object);
	}
      g_free (type);
    }
  g_free (key);
  return TRUE;
}

static gboolean
rpd_remove_hash_foreach (gpointer key, gpointer value, gpointer user_data)
{
  GHashTable *type = (GHashTable *) value;
  if (type)
    {
      g_hash_table_foreach_remove (type, rpd_remove_field_foreach, NULL);
      g_hash_table_destroy (type);
    }

  g_free (key);
  return TRUE;
}

void
repligard_done_schema (RSchemaCfg * schemacfg)
{
  g_hash_table_foreach_remove (schemacfg->schema, rpd_remove_type_foreach,
			       NULL);
  g_hash_table_destroy (schemacfg->schema);
  g_hash_table_foreach_remove (schemacfg->links, rpd_remove_hash_foreach,
			       NULL);
  g_hash_table_destroy (schemacfg->links);
  g_free (schemacfg);
}

RSchemaField *
repligard_get_qualified_field (GHashTable * schema,
			       const gchar * object, const gchar * field)
{
  RSchemaType *table;
  g_assert ((schema != NULL) && (object != NULL) && (field != NULL));
  table = (RSchemaType *) g_hash_table_lookup (schema, object);
  if (table != NULL)
    return (RSchemaField *) g_hash_table_lookup (table->object, field);
  return NULL;
}

GHashTable *
repligard_get_object (GHashTable * schema, const gchar * object)
{
  g_assert ((schema != NULL) && (object != NULL));
  return ((RSchemaType *) g_hash_table_lookup (schema, object))->object;
}

RSchemaType *
repligard_get_object_type (GHashTable * schema, const gchar * object)
{
  g_assert ((schema != NULL) && (object != NULL));
  return (RSchemaType *) g_hash_table_lookup (schema, object);
}

RSchemaField *
repligard_get_field (GHashTable * object, const gchar * field)
{
  g_assert ((object != NULL) && (field != NULL));
  return (RSchemaField *) g_hash_table_lookup (object, field);
}
