/* $Id: midgard.c,v 1.42 2002/12/08 03:25:52 armand Exp $
 *
 * midgard-lib: database access for Midgard clients
 *
 * Copyright (C) 1999 Jukka Zitting <jukka.zitting@iki.fi>
 * Copyright (C) 2000 The Midgard Project ry
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "midgard/midgard.h"
#include "midgard/authfailure.h"
#include "midgard/internal.h"
#include "defaults.h"
#include "midgard/parser.h"
#include "midgard/parsers.h"
#include "md5.h"
#if HAVE_CRYPT_H
#include <crypt.h>
#else
#define _XOPEN_SOURCE
#ifndef WIN32
#include <unistd.h>
#endif
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <signal.h>

#ifdef WIN32
#include <process.h>
#include "win32/flock.h"
#include "win32/win95nt.h"
#else
#include <sys/utsname.h>
#include <sys/file.h>
#include <netdb.h>
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 255
#endif

gchar *repligard_magic_string;
static FILE *_log_file = NULL;

void mgd_log_debug_none(const gchar *domain, GLogLevelFlags level,
      const gchar *msg, gpointer userdata)
{
}

void mgd_log_debug_default(const gchar *domain, GLogLevelFlags level,
      const gchar *msg, gpointer userdata)
{
   int fno;
   
   if (_log_file == NULL) _log_file = stderr;
   fno = fileno(_log_file);

   if (fno != 2) flock(fno, LOCK_EX);

   fprintf(_log_file, "%s:", domain != NULL ? domain : "midgard-lib");
   fprintf(_log_file, "%ld:", (unsigned long)getpid());

   switch (level) {
      case G_LOG_FLAG_RECURSION:
         fprintf(_log_file, "RECURSION");
         break;

      case G_LOG_FLAG_FATAL:
         fprintf(_log_file, "FATAL");
         break;

      case G_LOG_LEVEL_ERROR:
         fprintf(_log_file, "error");
         break;

      case G_LOG_LEVEL_CRITICAL:
         fprintf(_log_file, "CRITICAL");
         break;

      case G_LOG_LEVEL_WARNING:
         fprintf(_log_file, "warning");
         break;

      case G_LOG_LEVEL_MESSAGE:
         fprintf(_log_file, "message");
         break;

      case G_LOG_LEVEL_INFO:
         fprintf(_log_file, "info");
         break;

      case G_LOG_LEVEL_DEBUG:
         fprintf(_log_file, "debug");
         break;

      default:
         fprintf(_log_file, "UNKNOWN(%d)", level);
         break;
   }

   fprintf(_log_file, ":%s\n", msg);

   fflush(_log_file);

   if (fno != 2) flock(fno, LOCK_UN);
}

void mgd_init()
{
   mgd_init_ex(0, NULL);
}

guint mgd_parse_log_levels(const char *levels)
{
   const char *c;
   guint mask = 0;
   int level = 0;

   for (c = levels; *c != '\0'; ) {
      /* skip over non-text */
      if (!isalpha(*c)) { c++; continue; }

      /* first char determines log level */
      switch (tolower(*c)) {
         /*[eeh]  I don't think we're supposed to offer these as log
          * level flags
            case 'r': level = G_LOG_FLAG_RECURSION; break;
            case 'f': level = G_LOG_FLAG_FATAL; break;
         */
         case 'e': level = G_LOG_LEVEL_ERROR; break;
         case 'c': level = G_LOG_LEVEL_CRITICAL; break;
         case 'w': level = G_LOG_LEVEL_WARNING; break;
         case 'm': level = G_LOG_LEVEL_MESSAGE; break;
         case 'i': level = G_LOG_LEVEL_INFO; break;
         case 'd': level = G_LOG_LEVEL_DEBUG; break;
         default: level = 0; break;
      }

      /* skip over rest of 'word' */
      for (c++; *c != '\0' && isalpha(*c); c++) {}

      /* last word was garbage, ignore, maybe we should log this */
      if (level == 0) continue;

      /* select single log level */
      if (*c != '+') {
         mask |= level;
         continue;
      }

      /* select loglevel and all higher loglevels */
      for (; level && level > G_LOG_FLAG_FATAL; level >>= 1) {
         mask |= level;
      }
      c++;
   }

   return mask;
}

void mgd_init_ex(guint log_levels, char *logname)
{
	if (mgd_parser_list())
		return;
	(void) mgd_parser_create("latin1", "ISO-8859-1", 1);
	(void) mgd_parser_init_russian();
	{
#ifdef WIN32
		/* WinSock 2 params for resolving the hostname */
		char str[MAXHOSTNAMELEN + 7];
		char hostname[MAXHOSTNAMELEN + 1] = "\0";
		struct hostent *p;
	    WSADATA ver;
#else
		struct hostent *hs;
	    char hostname[255];
		struct utsname data;
#endif
	    gchar *hsaddr = NULL;
	    hostname[0]=0;
#ifdef WIN32
		WSAStartup(MAKEWORD(2, 0), &ver);
		if ((gethostname(str, MAXHOSTNAMELEN) != 0)) {
			p = gethostbyname(str);
			WSACleanup();
		if(p) hsaddr = g_strdup(p->h_addr);
#else
	    if(gethostname(hostname, 254)) {
		hs = gethostbyname(hostname);
		endhostent();
#endif
	    } else {
                /*[eeh] not sure gethostname doesn't touch the
                 * parameter if it returns false */
	        hostname[0]=0;
            }
	    if(hsaddr == NULL) {
		hsaddr = g_strdup_printf("127.0.0.%d", rand()%255);
	    }
#ifdef WIN32
		/*
			<dsr@best-off.org>
			Build the repligard_magic_string on Win32
			We defnitly need more random data for this, but for
			testing purposes this one works *sigh*
		 */
		repligard_magic_string =
		    g_strdup_printf("%s|%s",hostname, hsaddr);
#else
		if(uname(&data) == 0) {
		repligard_magic_string = 
		    g_strdup_printf("%s|%s|%s|%s|%s|%s|%s",
			data.sysname,data.nodename,hostname,
			data.release,data.version,data.machine,
			hsaddr);
	    } else {
		repligard_magic_string =
		    g_strdup_printf("%s|%s",hostname, hsaddr);
	    }
#endif
		g_free(hsaddr);
	}

   if (log_levels != 0)
      g_log_set_handler("midgard-lib", log_levels, mgd_log_debug_default, NULL);
   if (~log_levels != 0)
      g_log_set_handler("midgard-lib", ~log_levels, mgd_log_debug_none, NULL);

   if (logname == NULL) {
      _log_file = stderr;
   } else {
      _log_file = fopen(logname, "a");
      if (_log_file == NULL) {
         _log_file = stderr;
         g_log("midgard-lib", G_LOG_LEVEL_WARNING,
               "Could not open logfile '%s', using stderr", logname);
      }
   }
}

void mgd_done()
{
	mgd_parser_free_all();
	g_free(repligard_magic_string);
	repligard_magic_string = NULL;
}

const char *mgd_version()
{
	return MIDGARD_LIB_VERSION;
}

midgard *mgd_setup()
{
	midgard *mgd;

	/* create midgard handle */
	mgd = (midgard *) malloc(sizeof (struct midgard));
	if (!mgd)
		return NULL;

	mgd->res = NULL;
	mgd->pool = mgd_alloc_pool();
	if (!mgd->pool) {
		free(mgd);
		return NULL;
	}
	mgd->tmp = mgd_alloc_pool();
	if (!mgd->tmp) {
		mgd_free_pool(mgd->pool);
		free(mgd);
		return NULL;
	}

   mgd->current_user = &mgd->user_at_pagestart;

   mgd->user_at_pagestart.id = 0;
   mgd->setuid_user.id = 0;
   mgd->user_at_pagestart.is_admin = 0;
   mgd->setuid_user.is_admin = 0;
#if HAVE_MIDGARD_SITEGROUPS
   mgd->user_at_pagestart.is_root = 0;
   mgd->setuid_user.is_root = 0;
   mgd->user_at_pagestart.sitegroup = 0;
   mgd->setuid_user.sitegroup = 0;
#endif
   mgd->user_at_pagestart.member_of = NULL;
   mgd->setuid_user.member_of = NULL;

	mgd->username = NULL;

	mgd->parser = mgd_parser_list();

	mgd->mysql = NULL;
   mgd->blobdir = NULL;

	return mgd;
}

void mgd_easy_connect(midgard *mgd, const char *host, const char *database,
      const char *user, const char *password)
{
int connected;

	/* create mysql connection handle */
	mgd->mysql = mysql_init(NULL);

   /* connect to the database */
   connected = mgd->mysql &&
      mysql_real_connect(mgd->mysql, host,
         user, password, database, 0, NULL, 0);

	if (!mgd->mysql || !connected) {
      /* EEH/TODO: I know this is ugly but we need this info.
         There will be proper logging at some point
      */
      g_log("midgard-lib", G_LOG_LEVEL_WARNING,
         "midgard-lib: mgd_connect to %s@%s://%s failed: %s",
            user, host, database,
            mgd->mysql ? mysql_error(mgd->mysql) : "<mysql server not found?>");

		if (mgd->mysql) mysql_close(mgd->mysql);
      mgd->mysql = NULL;
   }
}

midgard *mgd_connect(const char *hostname, const char *database,
		     const char *username, const char *password)
{
	midgard *mgd = mgd_setup();

	if (mgd == NULL)
		return NULL;

	/* use default values if none specified */
	if (!hostname)
		hostname = MGD_MYSQL_HOST;
	if (!username)
		username = MGD_MYSQL_USERNAME;
	if (!password)
		password = MGD_MYSQL_PASSWORD;
	if (!database)
		database = MGD_MYSQL_DATABASE;

	/* create mysql connection handle */
	mgd_easy_connect(mgd, hostname, database, username, password);

	if (!mgd->mysql) {
		mgd_free_pool(mgd->tmp);
		mgd_free_pool(mgd->pool);
		free(mgd);
		return NULL;
	}

	return mgd;
}

void mgd_close(midgard * mgd)
{
	assert(mgd);
	mgd_clear(mgd, MGD_CLEAR_ALL);

	if (mgd->mysql) {
		mysql_close(mgd->mysql);
      mgd->mysql = NULL;
	}

   if (mgd->user_at_pagestart.member_of != NULL)
      free(mgd->user_at_pagestart.member_of);
   if (mgd->setuid_user.member_of != NULL)
      free(mgd->setuid_user.member_of);

	mgd_free_pool(mgd->tmp);
	mgd_free_pool(mgd->pool);
	free(mgd);
}

int mgd_select_db(midgard * mgd, const char *database)
{
int errcode;

   assert(mgd);
   assert(database);

	if ((errcode = mysql_select_db(mgd->mysql, database)) == 0) return 1;

   g_log("midgard-lib", G_LOG_LEVEL_ERROR,
         "Midgard: mgd_select_db(%s) failed: %s (%d)\n",
		      database, mysql_error(mgd->mysql), errcode);
	return 0;
}

int mgd_assert_db_connection(midgard *mgd, const char *hostname, const char *database,
      const char* username, const char *pass)
{
void (*handler) (int);

   handler=signal(SIGPIPE, SIG_IGN);

   /* EEH: ensure that the link did not die, from PHPs mysql connector */
#if defined(CR_SERVER_GONE_ERROR)
   mysql_stat(mgd->mysql);
   if (mysql_errno(mgd->mysql) == CR_SERVER_GONE_ERROR) {
#else
   if (!strcasecmp(mysql_stat(mgd->mysql), "mysql server has gone away")) {
#endif
		mysql_close(mgd->mysql);
      sleep(1);
      mgd_easy_connect(mgd, hostname, database, username, pass);
   }

   signal(SIGPIPE, handler);
   return (mgd->mysql != NULL);
}

#define MGD_USERNAME_MAXLEN 80

static int mgd_auth_su(midgard * mgd, const char *username,
		       const char *password, int su, int setuid)
{
	midgard_res *res;
	int i;
	const char *cipher, *passwd;
#if HAVE_MIDGARD_SITEGROUPS
	int host_sitegroup, req_sitegroup, admingroup;
	char *sitegroup;
	int rootuser = 0;
	char delim = '+';
	int grant;
	char *reauth = NULL;
	char uname[MGD_USERNAME_MAXLEN + 1];
#endif

	assert(mgd);

	if (mgd->mysql == NULL)
		return MGD_AUTH_NOT_CONNECTED;

#if HAVE_MIDGARD_SITEGROUPS
	host_sitegroup = mgd->current_user->sitegroup;
#endif
	mgd_clear(mgd, MGD_CLEAR_BASE);

   if (setuid) {
      mgd->current_user = &mgd->setuid_user;
	   mgd_clear(mgd, MGD_CLEAR_BASE);
#if HAVE_MIDGARD_SITEGROUPS
      mgd->current_user->sitegroup = host_sitegroup;
#endif
   }

	/* anonymous */
	if (!username || !*username) {
		return MGD_AUTH_ANONYMOUS;
   }

	if (strlen(username) > MGD_USERNAME_MAXLEN)
		return MGD_AUTH_INVALID_NAME;

#if HAVE_MIDGARD_SITEGROUPS
#define MIDGARD_LOGIN_SG_ROOT_SEP "*!$"

#define MIDGARD_LOGIN_SG_ROOT_ROOT	'*'
#define MIDGARD_LOGIN_SG_ROOT_ADMIN	'!'
#define MIDGARD_LOGIN_SG_ROOT_USER	'$'
#define MIDGARD_LOGIN_SG_ADMIN_USER	';'
#define MIDGARD_LOGIN_SG_USER_USER	'+'
#define MIDGARD_LOGIN_SG_REAUTH		'='

	strcpy(uname, username);

	sitegroup = strpbrk(uname, MIDGARD_LOGIN_SG_SEPARATOR);

   if (sitegroup) {
      delim = *sitegroup;
      *sitegroup = '\0';
      sitegroup++;
      rootuser = (strchr(MIDGARD_LOGIN_SG_ROOT_SEP, delim) != NULL);

      if (rootuser && !*sitegroup) return MGD_AUTH_SG_NOTFOUND;
   } else if (host_sitegroup == 0) {
      delim = MIDGARD_LOGIN_SG_ROOT_ROOT;
      rootuser = 1;
   } else {
      delim = MIDGARD_LOGIN_SG_USER_USER;
      rootuser = 0;
   }

	reauth = strchr(uname, MIDGARD_LOGIN_SG_REAUTH);
	if (reauth) {
		*reauth = '\0';
		reauth++;
	}

	if (su && (sitegroup || reauth)) {
		return MGD_AUTH_REAUTH;
	}

	if (sitegroup && *sitegroup) {
		res =
		   mgd_ungrouped_select(mgd, "id,admingroup", "sitegroup",
					"name=$q", NULL, sitegroup);

		/* sitegroup must be found */
		if (!res || !mgd_fetch(res)) {
			if (res)
				mgd_release(res);
			return MGD_AUTH_SG_NOTFOUND;
		}

		req_sitegroup = mgd_sql2id(res, 0);
		admingroup = mgd_sql2id(res, 1);
		mgd_release(res);
	}
	else {
		req_sitegroup = host_sitegroup;
		admingroup = 0;

		res =
		   mgd_ungrouped_select(mgd, "admingroup", "sitegroup",
					"id=$d", NULL, host_sitegroup);

		if (res) {
			if (mgd_fetch(res))
				admingroup = mgd_sql2id(res, 0);
			mgd_release(res);
		}
	}
#endif

#if HAVE_MIDGARD_SITEGROUPS
	if (rootuser) {
		res = mgd_ungrouped_select(mgd, "person.id,person.password",
					   "person,member",
					   "person.username=$q AND person.sitegroup=0"
					   " AND member.uid = person.id AND member.gid=0",
					   NULL, uname);
	}
	else {
		res = mgd_ungrouped_select(mgd, "id,password", "person",
					   "username=$q AND sitegroup=$d",
					   NULL, uname, req_sitegroup);
	}

	/* username not found */
	if (!res) {
		return MGD_AUTH_NOTFOUND;
	}

	/* only one user may match */
	if (mgd_rows(res) != 1) {
      mgd_release(res);
		return MGD_AUTH_DUPLICATE;
	}
#else
	res = mgd_ungrouped_select(mgd, "id,password", "person", "username=$q",
				   NULL, username);
#endif

	/* username not found */
	if (!res) {
		return MGD_AUTH_NOTFOUND;
	}

	while (mgd->current_user->id == 0 && mgd_fetch(res)) {
		passwd = mgd_colvalue(res, 1);

		if (passwd[0] == '*' && passwd[1] == '*') {
			passwd += 2;
			cipher = password;
		}
		else
#ifdef WIN32
			g_log("midgard-lib", G_LOG_LEVEL_WARNING, "crypt not supported on Win32");
#else
			cipher = (const char *)crypt(password, passwd);
#endif

		if (
#if HAVE_MIDGARD_SITEGROUPS
			   su ||
#endif
			   !strcmp(cipher, passwd))
			mgd->current_user->id = mgd_sql2id(res, 0);
	}

	mgd_release(res);

	if (mgd->current_user->id == 0) {
		return MGD_AUTH_INVALID_PWD;
	}

	/* Get group information */
#if HAVE_MIDGARD_SITEGROUPS
	res = mgd_ungrouped_select(mgd, "gid", "member",
				   "uid=$i AND sitegroup in (0, $i)",
				   "gid DESC", mgd->current_user->id, req_sitegroup);
#else
	res =
	   mgd_ungrouped_select(mgd, "gid", "member", "uid=$i", "gid DESC",
				mgd->current_user->id);
#endif

	if (res) {
      mgd->current_user->member_of
         = (int *) realloc(mgd->current_user->member_of,
                           sizeof (int) * (mgd_rows(res) + 1));

		i = 0;
		while (mgd_fetch(res)) {
			if ((mgd->current_user->member_of[i] = mgd_sql2id(res, 0)) != 0) {
#if HAVE_MIDGARD_SITEGROUPS
				if (mgd->current_user->member_of[i] == admingroup)
					mgd->current_user->is_admin = 1;
#endif
				i++;
			}
			else {
            mgd->current_user->is_admin = 1;
#if HAVE_MIDGARD_SITEGROUPS
            mgd->current_user->is_root = 1;
#endif
			}
      }

		mgd->current_user->member_of[i] = 0;
		mgd_release(res);
	}

#if HAVE_MIDGARD_SITEGROUPS
	grant = (rootuser && mgd->current_user->is_root);
	if (!grant && !rootuser && !mgd->current_user->is_root) {
		if (host_sitegroup != 0 && req_sitegroup == host_sitegroup)
			grant = 1;
		if (host_sitegroup == 0 && req_sitegroup != 0)
			grant = 1;
	}

	if (!grant) {
		mgd_clear(mgd, MGD_CLEAR_BASE);
	}
	else {
		/* Change current sitegroup */
		mgd->current_user->sitegroup = req_sitegroup;
	}

	switch (delim) {
      case MIDGARD_LOGIN_SG_ROOT_USER:
      case MIDGARD_LOGIN_SG_ADMIN_USER:
         mgd->current_user->is_admin = 0;
      case MIDGARD_LOGIN_SG_ROOT_ADMIN:
         mgd->current_user->is_root = 0;
	}

	if (su) {
		if (!grant || mgd->current_user->sitegroup == 0
            || !mgd->current_user->is_admin || !reauth) {
			mgd_clear(mgd, MGD_CLEAR_BASE);
		}
		else {
			/* this depends on mgd->sitegroup already being set */
			mgd_auth_su(mgd, reauth, "", 1, 0);
		}
	}
#endif

	return mgd->current_user->id;
}

int mgd_auth(midgard * mgd,
      const char *username, const char *password, int setuid)
{
	return mgd_auth_su(mgd, username, password, 0, setuid);
}

int mgd_auth_is_setuid(midgard * mgd)
{
   return (mgd->current_user == &mgd->setuid_user);
}

MGD_API int mgd_auth_unsetuid(midgard * mgd)
{
   if (mgd->current_user == &mgd->user_at_pagestart) return 0;
   mgd->current_user = &mgd->user_at_pagestart;
   return 1;
}

void mgd_clear(midgard * mgd, int clearflags)
{
	assert(mgd);

	/* clear user information */
	mgd->current_user->id = 0;
   mgd->current_user->is_admin = 0;
#if HAVE_MIDGARD_SITEGROUPS
	mgd->current_user->is_root = 0;
   if (clearflags & MGD_CLEAR_SITEGROUP) { mgd->current_user->sitegroup = 0; }
#endif
	mgd->username = NULL;

	if (mgd->current_user->member_of != NULL)
	   free(mgd->current_user->member_of);
	mgd->current_user->member_of = NULL;

	/* we automaticaly select first parser to avoid confusion */
	if (clearflags & MGD_CLEAR_PARSER) { mgd->parser = mgd_parser_list(); }

	/* clear resources reserved */
	while (mgd->res) mgd_release(mgd->res);

	mgd_clear_pool(mgd->pool);
}

int mgd_errno(midgard * mgd)
{
	assert(mgd);
	return mysql_errno(mgd->mysql);
}

const char *mgd_error(midgard * mgd)
{
	assert(mgd);
	return mysql_error(mgd->mysql);
}

int mgd_select_parser(midgard * mgd, const char *name)
{
	return mgd_parser_activate(mgd, name);
}

const char *mgd_get_parser_name(midgard * mgd)
{
	if (mgd && mgd->parser)
		return mgd->parser->name;
	return NULL;
}

const char *mgd_get_encoding(midgard * mgd)
{
	if (mgd && mgd->parser)
		return mgd->parser->encoding;
	return NULL;
}

int mgd_mail_need_qp(midgard * mgd)
{
	if (mgd && mgd->parser)
		return mgd->parser->need_qp;
	return 0;
}

#if HAVE_MIDGARD_SITEGROUPS
void mgd_set_sitegroup(midgard * mgd, int sitegroup)
{
	assert(mgd);
	mgd->current_user->sitegroup = sitegroup;
}
#endif

#if HAVE_MIDGARD_SITEGROUPS
int mgd_sitegroup(midgard * mgd)
{
	assert(mgd);
	return mgd->current_user->sitegroup;
}
#endif

int mgd_user(midgard * mgd)
{
	assert(mgd);
	return mgd->current_user->id;
}

int mgd_isauser(midgard * mgd)
{
	assert(mgd);
	return (mgd->current_user->id != 0);
}

int mgd_isadmin(midgard * mgd)
{
	assert(mgd);
	return mgd->current_user->is_admin;
}

#if HAVE_MIDGARD_SITEGROUPS
int mgd_isroot(midgard * mgd)
{
	assert(mgd);
	return mgd->current_user->is_root;
}

void mgd_force_root(midgard * mgd)
{
	assert(mgd);
	mgd->current_user->is_root = 1;
	mgd->current_user->is_admin = 1;
	mgd->current_user->id = 0;
}
#endif

void mgd_force_admin(midgard * mgd)
{
	assert(mgd);
	mgd->current_user->is_admin = 1;
	mgd->current_user->id = 0;
}

int mgd_isuser(midgard * mgd, int user)
{
	assert(mgd);
	return (mgd->current_user->id == user);
}

int mgd_ismember(midgard * mgd, int group)
{
	int i;
   int *member_of;

	assert(mgd);

   member_of = mgd->current_user->member_of;

	if (member_of == NULL) return 0;

	for (i = 0; member_of[i]; i++) if (member_of[i] == group) return 1;

	return 0;
}

int *mgd_groups(midgard * mgd)
{
	return mgd->current_user->member_of;
}

midgard_res *mgd_query(midgard * mgd, const char *query, ...)
{
	midgard_res *res;
	va_list args;
	va_start(args, query);
	res = mgd_vquery(mgd, query, args);
	va_end(args);
	return res;
}

midgard_res *mgd_vquery(midgard * mgd, const char *query, va_list args)
{
	midgard_res *res;
	midgard_pool *pool;
	MYSQL_RES *mres;
	int rv;
	const char *fquery;
	assert(mgd);
	assert(query);

	if (mgd->mysql == NULL)
		return NULL;

	/* format and send query */
	pool = mgd_alloc_pool();
	if (!pool)
		return NULL;
	fquery = mgd_vformat(mgd, pool, query, args);
	if (!fquery) {
		mgd_free_pool(pool);
		return NULL;
	}

   g_log("midgard-lib", G_LOG_LEVEL_DEBUG, "query=%s", fquery);

	rv = mysql_query(mgd->mysql, fquery);
	mgd_free_pool(pool);

   if (rv != 0) {
      g_log("midgard-lib", G_LOG_LEVEL_DEBUG,
            "query failed: %s", mysql_error(mgd->mysql));
		return NULL;
   }

	/* get results */
	mres = mysql_store_result(mgd->mysql);
	if (!mres)
		return NULL;
	if (mysql_num_rows(mres) == 0) {
		mysql_free_result(mres);
		return NULL;
	}
	/* create result handle */
	res = (midgard_res *) malloc(sizeof (struct midgard_res));
	if (!res) {
		mysql_free_result(mres);
		return NULL;
	}
	res->pool = mgd_alloc_pool();
	if (!res->pool) {
		free(res);
		mysql_free_result(mres);
		return NULL;
	}
	res->mgd = mgd;
	res->res = mres;
	res->row = NULL;
	res->rown = 0;
	res->rows = mysql_num_rows(mres);
	res->cols = mysql_num_fields(mres);
   res->fields = mysql_fetch_fields(mres);

	/* insert into result chain */
	if (mgd->res)
		mgd->res->prev = res;
	res->prev = NULL;
	res->next = mgd->res;
	mgd->res = res;

	return res;
}

midgard_res *mgd_ungrouped_select(midgard * mgd,
				  const char *fields, const char *from,
				  const char *where, const char *sort, ...)
{
	midgard_res *res;
	va_list args;
	va_start(args, sort);
	res = mgd_ungrouped_vselect(mgd, fields, from, where, sort, args);
	va_end(args);
	return res;
}

#if HAVE_MIDGARD_SITEGROUPS
midgard_res *mgd_sitegroup_select(midgard * mgd,
				  const char *fields, const char *from,
				  const char *where, const char *sort, ...)
{
	midgard_res *res;
	va_list args;
	va_start(args, sort);
	res = mgd_sitegroup_vselect(mgd, fields, from, where, sort, args);
	va_end(args);
	return res;
}
#endif

midgard_res *mgd_ungrouped_vselect(midgard * mgd,
				   const char *fields, const char *from,
				   const char *where, const char *sort,
				   va_list args)
{
	midgard_res *res;
	const char *query;
	midgard_pool *pool;
	assert(fields && from);
	pool = mgd_alloc_pool();
	if (!pool)
		return NULL;
	/* format query */
	if (where)
		if (sort)
			query = mgd_format(mgd, pool,
					   "SELECT $s FROM $s WHERE $s ORDER BY $s",
					   fields, from, where, sort);
		else
			query =
			   mgd_format(mgd, pool, "SELECT $s FROM $s WHERE $s",
				      fields, from, where);
	else if (sort)
		query =
		   mgd_format(mgd, pool, "SELECT $s FROM $s ORDER BY $s",
			      fields, from, sort);
	else
		query =
		   mgd_format(mgd, pool, "SELECT $s FROM $s", fields, from);
	if (!query) {
		mgd_free_pool(pool);
		return NULL;
	}

	/* execute query */
	res = mgd_vquery(mgd, query, args);

	mgd_free_pool(pool);
	return res;
}

#if HAVE_MIDGARD_SITEGROUPS
/* EmHe: add support for named tables at some point */
char *mgd_sitegroup_clause(midgard * mgd, midgard_pool * pool,
			   const char *tables)
{
	char *table;
	const char *sgclause;

	if (mgd->current_user->is_root) {
		return "";
	}

	table = strtok(mgd_strdup(pool, tables), " ,");
	while (table && !*table)
		table = strtok(NULL, " ,");

	if (!table) {
		return "AND 0=1";
	}			/* EmHe: this is an error */

	sgclause = mgd_format(mgd, pool, "AND ($s.sitegroup in (0,$d)",
			      table, mgd->current_user->sitegroup);

	if (!sgclause) {
		return "AND 0=2";
	}			/* EmHe: this is an error */

	for (table = strtok(NULL, " ,"); table; table = strtok(NULL, " ,")) {
		if (!*table)
			continue;

		sgclause =
		   mgd_format(mgd, pool, "$s AND $s.sitegroup in (0, $d)",
			      sgclause, table, mgd->current_user->sitegroup);

		if (!sgclause) {
			return "AND 0=3";
		}		/* EmHe: this is an error */
	}

	return mgd_format(mgd, pool, "$s)", sgclause);
}
#endif

#if HAVE_MIDGARD_SITEGROUPS
midgard_res *mgd_sitegroup_vselect(midgard * mgd,
				   const char *fields, const char *from,
				   const char *where, const char *sort,
				   va_list args)
{
	midgard_res *res;
	const char *query;
	midgard_pool *pool;
	const char *sgclause;

	assert(fields && from);

	pool = mgd_alloc_pool();
	if (!pool)
		return NULL;

	sgclause = mgd_sitegroup_clause(mgd, pool, from);

	if (sort)
		query = mgd_format(mgd, pool,
				   "SELECT $s FROM $s WHERE ($s) $s"
				   " ORDER BY $s",
				   fields, from, where ? where : "1=1",
				   sgclause, sort);
	else
		query = mgd_format(mgd, pool,
				   "SELECT $s FROM $s WHERE ($s) $s",
				   fields, from, where ? where : "1=1",
				   sgclause);

	if (!query) {
		mgd_free_pool(pool);
		return NULL;
	}

	/* execute query */
	res = mgd_vquery(mgd, query, args);

	mgd_free_pool(pool);
	return res;
}
#endif

midgard_res *mgd_ungrouped_record(midgard * mgd,
				  const char *fields, const char *table, int id)
{
	midgard_res *res;
	const char *query;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return NULL;

	/* format query */
	query = mgd_format(mgd, pool, "SELECT $s FROM $s WHERE id=$i",
			   fields, table, id);

	if (!query) {
		mgd_free_pool(pool);
		return NULL;
	}

	/* execute query */
	res = mgd_query(mgd, query);

	mgd_free_pool(pool);
	return res;
}

#if HAVE_MIDGARD_SITEGROUPS
midgard_res *mgd_sitegroup_record(midgard * mgd,
				  const char *fields, const char *table, int id)
{
	midgard_res *res;
	const char *query;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return NULL;

	/* format query */
	query = mgd_format(mgd, pool,
			   "SELECT $s,sitegroup FROM $s WHERE id=$i AND (sitegroup in (0, $d) OR $d<>0)",
			   fields, table, id,
            mgd->current_user->sitegroup,
            mgd->current_user->is_root);
	if (!query) {
		mgd_free_pool(pool);
		return NULL;
	}

	/* execute query */
	res = mgd_query(mgd, query);

	mgd_free_pool(pool);
	return res;
}
#endif

int mgd_idfield(midgard * mgd, const char *field, const char *table, int id)
{
	midgard_res *res;
	int rv;

	/* get record field */
	res = mgd_ungrouped_record(mgd, field, table, id);
	if (!res)
		return 0;

	/* get id value */
	if (mgd_fetch(res))
		rv = mgd_sql2id(res, 0);
	else
		rv = 0;
	mgd_release(res);

	return rv;
}

static char *midgard_all_tables[] = {
	"article",
	"element",
	"event",
	"eventmember",
	"file",
	"grp",
	"host",
	"image",
	"member",
	"page",
	"pageelement",
	"pagelink",
	"person",
	"preference",
	"snippet",
	"snippetdir",
	"style",
	"topic",
	NULL
	   /* "sitegroup" omitted because it's more of a meta-info kind of record */
	   /* "repligard" omitted because it's more of a meta-info kind of record */
};

int mgd_global_exists(midgard * mgd, const char *where, ...)
{
	midgard_pool *pool;
	char **from = &midgard_all_tables[0];
	const char *query;
	va_list args;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;

	va_start(args, where);
	query = mgd_vformat(mgd, pool, where, args);
	va_end(args);

	while (*from) {
		if (mgd_exists_id(mgd, *from, query)) {
			mgd_free_pool(pool);
			return 1;
		}

		from++;
	}

	mgd_free_pool(pool);
	return 0;
}

int mgd_exists_id(midgard * mgd, const char *from, const char *where, ...)
{
	midgard_res *res;
	const char *query;
	va_list args;
	int rv = 0;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;

	/* format query */
	if (strchr(from, ',')) return 0;

   query =
	   mgd_format(mgd, pool, "SELECT id FROM $s WHERE ($s)"
#if HAVE_MIDGARD_SITEGROUPS
      " AND (sitegroup IN (0,$d) OR $d<>0)"
#endif
      , from, where
#if HAVE_MIDGARD_SITEGROUPS
      , mgd->current_user->sitegroup, mgd->current_user->is_root
#endif
      );

	if (!query) {
		mgd_free_pool(pool);
		return 0;
	}

	/* execute query */
	va_start(args, where);
	res = mgd_vquery(mgd, query, args);
	va_end(args);

	mgd_free_pool(pool);
	if (!res)
		return 0;
	if (mgd_fetch(res))
		rv = atoi(mgd_colvalue(res, 0));

	mgd_release(res);
	return rv;
}

int mgd_exists_bool(midgard * mgd, const char *from, const char *where, ...)
{
	midgard_res *res;
	const char *query;
	va_list args;
	int rv = 0;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;

	query =
	   mgd_format(mgd, pool, "SELECT 1 FROM $s WHERE $s", from,
		      where);

	if (!query) {
		mgd_free_pool(pool);
		return 0;
	}

	/* execute query */
	va_start(args, where);
	res = mgd_vquery(mgd, query, args);
	va_end(args);

	mgd_free_pool(pool);
	if (!res)
		return 0;
	if (mgd_fetch(res))
		rv = atoi(mgd_colvalue(res, 0));

	mgd_release(res);
	return rv;
}

int mgd_fetch(midgard_res * res)
{
	assert(res && res->res);
	if (res->row)
		res->rown++;
	res->row = mysql_fetch_row(res->res);
	if (!res->row)
		return 0;
	return 1;
}

void mgd_release(midgard_res * res)
{
	assert(res);

   if (res->res == NULL) {
      g_log("midgard-lib", G_LOG_LEVEL_WARNING, "Someone is holding a stale reference to a midgard handle");
      return;
   }

	mysql_free_result(res->res);
   res->res = NULL;
	mgd_free_pool(res->pool);
	if (res->next)
		res->next->prev = res->prev;
	if (res->prev)
		res->prev->next = res->next;
	else
		res->mgd->res = res->next;
	free(res);
}

int mgd_rows(midgard_res * res)
{
	assert(res);
	return res->rows;
}

int mgd_cols(midgard_res * res)
{
	assert(res);
	return res->cols;
}

const char *mgd_colname(midgard_res * res, int i)
{
	assert(res && res->res);
	if (0 <= i && i < res->cols)
#if MYSQL_VERSION_ID > 32225
		return mysql_fetch_field_direct(res->res, i)->name;
#else
		return mysql_fetch_field_direct(res->res, i).name;
#endif
	else
		return NULL;
}

int mgd_colisnum(midgard_res * res, int i)
{
	assert(res && res->res);
	if (0 < i && i >= res->cols) return 0;

   switch (res->fields[i].type) {
      case FIELD_TYPE_TINY:
      case FIELD_TYPE_SHORT:
      case FIELD_TYPE_LONG:
      case FIELD_TYPE_INT24:
      case FIELD_TYPE_LONGLONG:
         return 1;

      default:
         return 0;
   }
}

const char *mgd_colvalue(midgard_res * res, int i)
{
	assert(res && res->row);
	if (0 <= i && i < res->cols)
		return res->row[i];
	else
		return NULL;
}

const char *mgd_column(midgard_res * res, const char *name)
{
	int i;
	for (i = 0; i < mgd_cols(res); i++)
		if (strcmp(mgd_colname(res, i), name) == 0)
			return mgd_colvalue(res, i);
	return NULL;
}

int mgd_exec(midgard * mgd, const char *command, ...)
{
	int rv;
	va_list args;
	va_start(args, command);
	rv = mgd_vexec(mgd, command, args);
	va_end(args);
	return rv;
}

int mgd_vexec(midgard * mgd, const char *command, va_list args)
{
	int rv;
	char *fcommand;
	midgard_pool *pool;
	assert(mgd);
	assert(command);

	if (mgd->mysql == NULL)
		return 0;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;
	/* format and send command */
	fcommand = mgd_vformat(mgd, pool, command, args);
	if (!fcommand) {
		mgd_free_pool(pool);
		return 0;
	}
	rv = mysql_query(mgd->mysql, fcommand);

   g_log("midgard-lib", G_LOG_LEVEL_DEBUG, "query=%s", fcommand);

   if (rv != 0) {
      g_log("midgard-lib", G_LOG_LEVEL_DEBUG,
            "query failed: %s", mysql_error(mgd->mysql));
   }

	mgd_free_pool(pool);
	return rv ? 0 : 1;
}

int mgd_create(midgard * mgd, const char *table,
	       const char *fields, const char *values, ...)
{
	int id;
	va_list args;
	va_start(args, values);
	id = mgd_vcreate(mgd, table, fields, values, args);
	va_end(args);
	return id;
}


int mgd_vcreate(midgard * mgd, const char *table,
		const char *fields, const char *values, va_list args)
{
	const char *command = NULL;
	int rv;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;
	/* format command */
#if HAVE_MIDGARD_SITEGROUPS
	if ((strcmp(table, "sitegroup") == 0)
         || (mgd->current_user->is_root && mgd->current_user->sitegroup == 0)) {
		/* records on sitegroup table don't get assigned to a sitegroup */
		if (mgd->current_user->is_root)
			command = mgd_format(mgd, pool,
					     "INSERT INTO $s ($s) VALUES ($s)",
					     table, fields, values);
	}
	else if (strstr(fields, "sitegroup") != NULL)
	{
		/* sitegroup information provided in fields */
		if (mgd->current_user->is_root && mgd->current_user->sitegroup == 0)
		{
			command = mgd_format(mgd, pool, "INSERT INTO $s ($s) VALUES ($s)",
				table, fields, values);
		}
		else
		{
			/* no rights to specify the sitegroup field if
			 * not SG0 admin logged into SG0
			 */
			return 0;
		}
	}
	else
	{
		/* no sitegroup information provided */
		command = mgd_format(mgd, pool,
				     "INSERT INTO $s ($s, sitegroup) VALUES ($s, $d)",
				     table, fields, values, mgd->current_user->sitegroup);
	}
#else
	command = mgd_format(mgd, pool,
			     "INSERT INTO $s ($s) VALUES ($s)",
			     table, fields, values);
#endif
	if (!command) {
		mgd_free_pool(pool);
		return 0;
	}

	/* execute command */
	rv = mgd_vexec(mgd, command, args);

	mgd_free_pool(pool);
	return rv ? mysql_insert_id(mgd->mysql) : 0;
}

int mgd_update(midgard * mgd, const char *table, int id,
	       const char *fields, ...)
{
	int rv;
	va_list args;
	va_start(args, fields);
	rv = mgd_vupdate(mgd, table, id, fields, args);
	va_end(args);
	return rv;
}

int mgd_vupdate(midgard * mgd, const char *table, int id,
		const char *fields, va_list args)
{
	const char *command = NULL;
	int rv;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;
	/* format command */
#if HAVE_MIDGARD_SITEGROUPS
	/* strcmp tolerable since updates are less common than gets */
	if (strcmp(table, "sitegroup") == 0) {
		if (mgd->current_user->is_root)
			command = mgd_format(mgd, pool,
					     "UPDATE $s SET $s WHERE id=$d",
					     table, fields, id);
	}
	else
		command = mgd_format(mgd, pool,
				     "UPDATE $s SET $s WHERE id=$d"
				     "  AND (sitegroup = $d OR $d<>0)",
				     table, fields, id, mgd->current_user->sitegroup,
				     mgd->current_user->is_root);
#else
	command = mgd_format(mgd, pool,
			     "UPDATE $s SET $s WHERE id=$d", table, fields, id);
#endif
	if (!command) {
		mgd_free_pool(pool);
		return 0;
	}

	/* execute command */
	rv = mgd_vexec(mgd, command, args);

	mgd_free_pool(pool);
	return rv;
}

int mgd_delete(midgard * mgd, const char *table, int id)
{
	const char *command = NULL;
	int rv;
	midgard_pool *pool;

	pool = mgd_alloc_pool();
	if (!pool)
		return 0;
	/* format command */
#if HAVE_MIDGARD_SITEGROUPS
	/* strcmp tolerable since updates are less common than gets */
	if (strcmp(table, "sitegroup") == 0) {
		if (mgd->current_user->is_root)
			command = mgd_format(mgd, pool,
					     "DELETE FROM $s WHERE id=$d",
					     table, id);
	}
	else
		command = mgd_format(mgd, pool, "DELETE FROM $s WHERE id=$d"
				     " AND (sitegroup = $d OR $d<>0)",
				     table, id,
                 mgd->current_user->sitegroup,
                 mgd->current_user->is_root);
#else
	command =
	   mgd_format(mgd, pool, "DELETE FROM $s WHERE id=$d", table, id);
#endif
	if (!command) {
		mgd_free_pool(pool);
		return 0;
	}

	/* execute command */
	rv = mgd_exec(mgd, command);

	mgd_free_pool(pool);
	return rv;
}

void mgd_set_blobdir(midgard *mgd, char *blobdir)
{
   if (blobdir == NULL) mgd->blobdir = NULL;
   else mgd->blobdir = blobdir;
}

const char* mgd_get_blobdir(midgard *mgd)
{
   return (const char*)mgd->blobdir;
}

int midgard_lib_compiled_with_sitegroups()
{
#if HAVE_MIDGARD_SITEGROUPS
   return 1;
#else
   return 0;
#endif
}

#if MIDGARD_PHP_REQUEST_CONFIG_BUG_WORKAROUND
static void *hack_rcfg;
static void *hack_dcfg;

void mgd_php_bug_workaround_set_rcfg_dcfg(void *rcfg, void *dcfg)
{
	hack_rcfg = rcfg;
	hack_dcfg = dcfg;
}

void* mgd_php_bug_workaround_get_rcfg()
{
	return hack_rcfg;
}

void* mgd_php_bug_workaround_get_dcfg()
{
	return hack_dcfg;
}
#endif

