#include "midgard/internal.h"
#include "midgard/midgard.h"
#include "midgard/pageresolve.h"
#include <string.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <glib.h>
#include <stdio.h>

int mgd_find_host(midgard *mgd,
   const char *hostname, int port, const char *uri, int null_prefix,
   mgd_host_t *parsed)
{
midgard_res *res;

   parsed->host = 0;
   parsed->style = 0;
   parsed->page = 0;
   parsed->sitegroup = 0;
   parsed->auth_required = 0;
   parsed->prefix_length = 0;
   parsed->found = 0;

	res = mgd_ungrouped_select(mgd,
				   "id,style,root,info&1,Length(prefix)"
#if HAVE_MIDGARD_SITEGROUPS
				   ",sitegroup"
#endif
				   , "host",

               "name=$q AND port in (0,$d) AND"
               " ( ($d <> 0 AND prefix = '')"
               "  OR ($d = 0 AND prefix=Left($q,Length(prefix)) AND"
               "      IF($d>=Length(prefix),MID($q,1+Length(prefix),1) IN ('','/'),0)=1"
               "     )"
               " )"
               " AND online <> 0",

				   "prefix DESC, port DESC",

               hostname, port, null_prefix, null_prefix, uri, strlen(uri), uri
				   );

	if (!res || mgd_rows(res) == 0 || !mgd_fetch(res)) {
      if (res) mgd_release(res);
		return 0;
   }

   parsed->host = atoi(mgd_colvalue(res, 0));
   parsed->style = atoi(mgd_colvalue(res, 1));
   parsed->page = atoi(mgd_colvalue(res, 2));
   parsed->auth_required = atoi(mgd_colvalue(res, 3));
   parsed->prefix_length = atoi(mgd_colvalue(res, 4));
#if HAVE_MIDGARD_SITEGROUPS
   parsed->sitegroup = atoi(mgd_colvalue(res, 5));
#else
   parsed->sitegroup = 0;
#endif

	mgd_release(res);
   parsed->found = 1;
   return 1;
}

void mgd_collect_page_elements(midgard *mgd, long page, int inherited_only,
   GHashTable *element_cache,
   mgd_store_style_elt_cb callback, void *userdata)
{
midgard_res *res;
const char *name;
const char *value;
long id;

   if (callback == NULL) return;

   res = mgd_ungrouped_select(mgd, "id,name,value", "pageelement", "page=$d $s",
      NULL,
      page, inherited_only ? "AND info&1<>0" : "");

   if (!res) return;

   while (mgd_fetch(res)) {
      id = atol(mgd_colvalue(res, 0));
      name = mgd_colvalue(res, 1);
      value = mgd_colvalue(res, 2);

      /* skip storing the element if it's allready defined */
      if (g_hash_table_lookup(element_cache, name) != NULL) { continue; }

      g_hash_table_insert(element_cache, g_strdup(name), "");
      mgd_cache_add(mgd, page, MGD_CACHE_ELT_PAGE, id);

      callback(name, value, userdata);
   }

   mgd_release(res);
}

void mgd_style_load(midgard *mgd, long cache_for_page, long style,
   GHashTable *element_cache,
   mgd_store_style_elt_cb callback, void *userdata)
{
midgard_res *res;
const char *name;
const char *value;
long id;

   if (callback == NULL) return;

   while (style) {
      res = mgd_ungrouped_select(mgd, "id,name,value", "element", "style=$d", NULL, style);

      if (res != NULL) {
         while (mgd_fetch(res)) {
            id = atol(mgd_colvalue(res, 0));
            name = mgd_colvalue(res, 1);
            value = mgd_colvalue(res, 2);

            /* skip storing the element if it's allready defined */
            if (g_hash_table_lookup(element_cache, name) != NULL) { continue; }

            g_hash_table_insert(element_cache, g_strdup(name), "");
            mgd_cache_add(mgd, cache_for_page, MGD_CACHE_ELT_STYLE, id);

            callback(name, value, userdata);
         }

         mgd_release(res);
      }

      style = mgd_idfield(mgd, "up", "style", style);
   }
}

int mgd_find_blob(midgard *mgd, mgd_page_t *page, char *blobname)
{
midgard_res *res;

   if (blobname == NULL) return 0;
   if (blobname[0] == '\0') return 0;
   if (strchr(blobname, '/') != NULL) return 0;

   res = mgd_ungrouped_select(mgd, "id,location", "blobs",
      "pid=$d AND ptable='page' AND name=$q", NULL,
      page->page, blobname);

   if (!res || !mgd_fetch(res)) {
      if (res) mgd_release(res);
      return 0;
   }

   blobname = (char*)mgd_colvalue(res,1);
   if (blobname != NULL && blobname[0] == '\0') { blobname = NULL; }
   if (blobname != NULL) { page->blob = atoi(mgd_colvalue(res, 0)); }
   mgd_release(res);
   page->found = MGD_FOUND_BLOB;
   return 1;
}

int mgd_parse_uri(midgard *mgd, mgd_host_t *host, char *uri,
   mgd_page_t *result,
   mgd_pathinfo_cb pathinfo_cb,
   GArray *path,
   void *userdata)
{
midgard_res *res;
char *tmp_uri;
char *basename, *ext;
int uri_len;
char *dir;
long tmp;
char *blob = NULL;

   result->style = host->style;
   result->active = 0;
   result->page = host->page;
   result->auth_required = host->auth_required;
   result->author = 0;
   result->found = MGD_FOUND_NONE;
   result->blob = 0;

   g_array_set_size(path, 0);

   g_array_append_val(path, host->page);

   basename = strrchr(uri + host->prefix_length, '/');
   ext = (basename != NULL) ? strchr(basename + 1, '.') : NULL;
   result->addslash = (basename == NULL)
                     || (basename[1] != '\0' && ext == NULL);

   if (ext && strcmp(ext, ".html") == 0) {
      uri_len = ext - uri - host->prefix_length;
      tmp_uri = mgd_strndup(mgd->pool, uri + host->prefix_length, uri_len);
   } else {
      tmp_uri = mgd_strdup(mgd->pool, uri + host->prefix_length);
      uri_len = strlen(tmp_uri);
   }

   if (!result->auth_required)
      result->auth_required = mgd_idfield(mgd, "info&1", "page", host->page);
   result->active = mgd_idfield(mgd, "info&2", "page", host->page);
   tmp = mgd_idfield(mgd, "style", "page", host->page);
   if (tmp != 0) result->style = tmp;

   /*[eeh]  We always have a root page */
   result->found = MGD_FOUND_PAGE;
   for (dir = strtok(tmp_uri, "/");
         dir != NULL;
         dir = strtok(NULL, "/")) {
      if (dir[0] == '\0') { continue; }

      res = mgd_ungrouped_select(mgd, "id,style,info&1,info&2,author", "page",
         "up=$d AND name=$q", NULL, result->page, dir);

      if (!res || !mgd_fetch(res)) {
         if (res) mgd_release(res);
         result->found = MGD_FOUND_NONE;
         break;
      }

      result->page = atoi(mgd_colvalue(res, 0));

      if ((tmp = atoi(mgd_colvalue(res, 1))) != 0) {
         result->style = tmp;
         /*[eeh]  don't reset path of pages so lower inherited page
          *       elements 'show through'
          */
      }

      if (!result->auth_required) result->auth_required = atoi(mgd_colvalue(res, 2));

      result->active = atoi(mgd_colvalue(res, 3));
      result->author = atol(mgd_colvalue(res, 4));

      g_array_append_val(path, result->page);

   }

   result->self_len = host->prefix_length + (dir != NULL ? (dir - tmp_uri) : uri_len);

   /*[eeh]  page not found or active page, try blob */
   if (result->found == MGD_FOUND_NONE || result->active) {
      blob = uri + result->self_len;
      if (mgd_find_blob(mgd, result, blob)) return 1;
   }

   if (result->found == MGD_FOUND_NONE && !result->active) {
      return 0;
   }

   if (result->active) {
      if (dir != NULL) result->addslash = 0;
      result->found = MGD_FOUND_PAGE;
      if (pathinfo_cb != NULL) {
         while (dir != NULL) {
            if (dir[0] != '\0') pathinfo_cb(dir, userdata);
            dir = strtok(NULL, "/");
         }
      }
   }

   return 1;
}

gboolean element_cache_free(gpointer key, gpointer value, gpointer userdata)
{
   g_free(key);
   return TRUE;
}

void mgd_load_styles(midgard *mgd, GArray* path, long style,
   mgd_store_style_elt_cb style_elt_cb, void *userdata,
   long cached_page)
{
int i;
midgard_res *res;
long page;
const char *title;
const char *content;
GHashTable *element_cache;
long parent;

   if (style_elt_cb == NULL || (cached_page == 0 && (path == NULL || path->len == 0))) return;

   if (cached_page != 0) { page = cached_page; }
   else { page = g_array_index(path, long, path->len - 1); }

   res = mgd_ungrouped_record(mgd, "title,content", "page", page);
   if (!res || !mgd_fetch(res)) { /* EEH: This should never happen */
      if (res) mgd_release(res);
      return;
   }

   element_cache = g_hash_table_new(g_str_hash, g_str_equal);

   title = mgd_colvalue(res, 0);
   content = mgd_colvalue(res, 1);
   style_elt_cb("title", title, userdata);
   style_elt_cb("content", content, userdata);

   style_elt_cb("title", title, userdata);
   g_hash_table_insert(element_cache, g_strdup("title"), "");
   style_elt_cb("content", content, userdata);
   g_hash_table_insert(element_cache, g_strdup("content"), "");

   if (cached_page != 0) {
      res = mgd_ungrouped_select(mgd,
         "element.name,element.value", "element,cache",
         "cache.page=$d AND cache.type=$d AND cache.id=element.id",
         NULL,
         cached_page, MGD_CACHE_ELT_STYLE);

      if (res) {
         while (mgd_fetch(res)) style_elt_cb(mgd_colvalue(res, 0), mgd_colvalue(res, 1), userdata);
         mgd_release(res);
      }

      res = mgd_ungrouped_select(mgd,
         "pageelement.name,pageelement.value", "pageelement,cache",
         "cache.page=$d AND cache.type=$d AND cache.id=pageelement.id",
         NULL,
         cached_page, MGD_CACHE_ELT_PAGE);

      if (res) {
         while (mgd_fetch(res)) style_elt_cb(mgd_colvalue(res, 0), mgd_colvalue(res, 1), userdata);
         mgd_release(res);
      }

      return;
   }

   mgd_collect_page_elements(mgd, page, 0, element_cache, style_elt_cb, userdata);

   while (parent = mgd_idfield(mgd, "up", "page", g_array_index(path, long, 0)))
   {
      g_array_prepend_val(path, parent);
   }

   for (i = path->len - 2; i >= 0; i--) {
      mgd_collect_page_elements(mgd, g_array_index(path, long, i), 1, element_cache, style_elt_cb, userdata);
   }

   mgd_style_load(mgd, page, style, element_cache, style_elt_cb, userdata);

   g_hash_table_foreach_remove(element_cache, element_cache_free, NULL);
   g_hash_table_destroy(element_cache);
}

void mgd_cache_touch(midgard *mgd, long page)
{
GString *command;
long sitegroup = 
#if MIDGARD_HAVE_SITEGROUPS
      (long)mgd_sitegroup(mgd)
#else
      (long)0
#endif
   ;

   if (sitegroup == 0) {
      mysql_query(mgd->mysql, "DELETE FROM cache");
   } else {
      command = g_string_new("");
      g_string_sprintf(command, "DELETE FROM cache WHERE sitegroup=%ld", sitegroup);
      mysql_query(mgd->mysql, command->str);
      g_string_free(command, TRUE);
   }
}

int mgd_cache_is_current(midgard *mgd, long page)
{
midgard_res *res;
int is_current;

   /*[eeh] page, type, id, sitegroup */
	res = mgd_ungrouped_select(mgd, "page", "cache", "page=$d LIMIT 1",
      NULL,
      page);

   is_current = (res != NULL && mgd_fetch(res));
   if (res != NULL) mgd_release(res);
   return is_current;
}

void mgd_cache_add(midgard *mgd, long page, int type, long id)
{
GString *command;

   command = g_string_new("");
   g_string_sprintf(command,
      "INSERT INTO cache (page,type,id,sitegroup) VALUES (%ld, %d, %ld, %ld)",
      page, type, id,
#if MIDGARD_HAVE_SITEGROUPS
      (long)mgd_sitegroup(mgd)
#else
      (long)0
#endif
      );
   mysql_query(mgd->mysql, command->str);
   g_string_free(command, TRUE);
}
