/* $Id: tree_core.c,v 1.12 2002/08/19 15:41:13 daniel Exp $
 *
 * midgard-lib: database access for Midgard clients
 *
 * Copyright (C) 1999 Jukka Zitting <jukka.zitting@iki.fi>
 * Copyright (C) 2000 The Midgard Project ry
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*! \file lib/src/tree_core.c
 *  \brief Tree core library
 *
 *  The functions of this file are used to manipulate the tree objects
 */


#include "midgard/midgard.h"
#include "midgard/internal.h"
#include "defaults.h"
#include "midgard/parser.h"
#include "midgard/parsers.h"
#if HAVE_CRYPT_H
#include <crypt.h>
#else
#define _XOPEN_SOURCE
#ifndef WIN32
#include <unistd.h>
#endif
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif

//! Initializer for a Midgard tree node
const tree_node zero_node = {
	0,
	0,
	0,
	0,
	NULL,
	NULL
}; 

//! Allocate and initialize a Midgard tree node
#define ALLOC_TNODE(pool, node) {\
	(node) = (tree_node*)mgd_alloc(pool, sizeof(tree_node));\
	*(node) = zero_node;\
}

/*! \brief Traverse the tree and execute a user defined function
 *         (please see mgd_walk_table_tree())
 *  \param mgd			Midgard handle
 *  \param tree			The tree to traverse
 *  \param order		Specify when the user function will be executed
 *  \param level		The current recursion level
 *  \param maxlevel		The maximum recursion level
 *  \param xparam		The user defined variable passed to the user
 *  					defined function
 *  \param func			The user defined function
 *  
 *  \return				non-0 if ok to continue, 0 to stop
 *
 *  walktree() may not be used outside of the lib.
 *
 */
static int walktree(midgard * mgd, tree_node * tree, int order, int level,
		    int maxlevel, void *xparam, midgard_userfunc func)
{
	int branch_ok = 1;
	tree_node *tmp;
	if(maxlevel && level > maxlevel) return 1;
	if (order && func)
		if(!(branch_ok = func(mgd, tree->id, level, xparam)))
			return 0;
	if(branch_ok > 0) {
		tmp = tree->child;
		while(tmp) {
			if(!walktree(mgd, tmp, order, level + 1, maxlevel, xparam, func))
				return 0;
			tmp = tmp->next;
		}
	}
	if ((!order) && func)
		if(!func(mgd, tree->id, level, xparam))
			return 0;
	return 1;
}

/*! \brief Get all the children of a list of nodes
 *  \param mgd			Midgard handle
 *  \param pool			A memory pool used during the process
 *  \param table		Identify the type of the node
 *  \param upfield		A string containing the name of the up field
 *  \param sort			The sort fields, separated by a comma
 *  \param ids			The ids of the father nodes
 *  \retval nodes		The children nodes
 *
 *  \return				The number of children
 *  
 *  mgd_tree_get_level() may not be used outside of the lib.
 *
 *  How this function work:<br>
 *  It gets the children and link them together according to their upfield,
 *  making them siblings.
 *  
 */
int mgd_tree_get_level(midgard *mgd, midgard_pool * pool, const char *table,
						const char * upfield, const char * sort,
						int ** ids, tree_node ** nodes)
{
	midgard_res *res;
	int i,n;

	if(!ids) return 0;
	if(!upfield || !(*upfield)) upfield = "up";
#if HAVE_MIDGARD_SITEGROUPS
	if((*ids)[0] && (*ids)[1]) {
		res = mgd_query(mgd,
			"SELECT id,$s FROM $s WHERE $s IN $D AND (sitegroup in (0, $d) OR $d<>0) ORDER BY $s,$s",
			upfield, table,
			upfield, *ids, mgd_sitegroup(mgd), mgd_isroot(mgd),
			upfield, sort ? sort : "id");
	} else { /* root query (single id) */
		res = mgd_query(mgd,
			"SELECT id,$s FROM $s WHERE $s=$d AND (sitegroup in (0, $d) OR $d<>0) ORDER BY $s,$s",
			upfield, table,
			upfield, (*ids)[0], mgd_sitegroup(mgd), mgd_isroot(mgd),
			upfield, sort ? sort : "id");
	}
#else
	if((*ids)[0] && (*ids)[1]) {
		res = mgd_query(mgd,
			"SELECT id,$s FROM $s WHERE $s IN $D ORDER BY $s, $s",
			upfield, table, upfield, *ids, upfield, sort ? sort : "id");
	} else { /* root query (single id) */
		res = mgd_query(mgd,
			"SELECT id,$s FROM $s WHERE $s=$d ORDER BY $s, $s",
			upfield, table, upfield, (*ids)[0], upfield, sort ? sort : "id");
	}
#endif
	if(!res || !(n = mgd_rows(res))) {
		nodes = NULL;
		ids = NULL;
		return 0;
	}
	*nodes = (tree_node *) mgd_alloc(pool, n * sizeof (tree_node));
	*ids = (int *) mgd_alloc(pool, (n + 1) * sizeof(int *));
	for(i = 0; mgd_fetch(res); i++) {
		(*ids)[i] = mgd_sql2int(res, 0);
		(*nodes)[i].id = (*ids)[i];
		(*nodes)[i].up = mgd_sql2int(res, 1);
		(*nodes)[i].level = 0;
		(*nodes)[i].size = 0;
		(*nodes)[i].child = NULL;
		(*nodes)[i].next = NULL;
		// if the current node and the previous node are from the same parent,
		// they are siblings so we update the next of the previous node
		if(i && (*nodes)[i].up == (*nodes)[i-1].up) {
			(*nodes)[i-1].next = &((*nodes)[i]);
		}
	}
	(*ids)[i] = 0;
	mgd_release(res);
	return n;
}

/*! \brief Extract all ids from a tree to build a flat tree ids list.
 *  \param tree			A subtree to process
 *  \param ids			The ids list
 *  \param index		An index on the tree list \a ids
 *  
 *  mgd_tree_get_ids() may not be used outside of the lib.
 *  \warning \a ids *must* be pre-allocated, for this, first call
 *  mgd_tree_init() on the tree, mgd_alloc(pool, tree->size + 1) and then you
 *  can call mgd_tree_get_ids().
 *  You have to set \a *index to 0 before calling this function.
 *  
 */
void mgd_tree_get_ids(tree_node * tree, int *ids, int *index) {
	tree_node * tmp;

	if(!tree) return;
	ids[*index] = tree->id;
	(*index)++;
	tmp = tree->child;
	while(tmp) {
		mgd_tree_get_ids(tmp, ids, index);
		tmp = tmp->next;
	}
}

/*! \brief Initialize internal parameters of the tree.
 *  \param tree			A subtree to process
 *  \param level		The recursion level
 *  
 *  mgd_tree_init() may not be used outside of the lib.
 *  
 */
int mgd_tree_init(tree_node * tree, int level) {
	tree_node * tmp;

	if(!tree) return 0;
	tree->size = 1;
	tree->level = level;
	tmp = tree->child;
	while(tmp) {
		tree->size += mgd_tree_init(tmp, level + 1);
		tmp = tmp->next;
	}
	
	return tree->size; /* store the size of the subtree + the current node */
}

/*! \brief Build a tree in memory
 *  \param mgd			Midgard handle
 *  \param pool			A memory pool used during the process
 *  \param table		Identify the type of the node
 *  \param upfield		A string containing the name of the up field
 *  \param root			The node id root of the subtree
 *  \param maxlevel		The maximum recursion depth
 *  \param sort			The sort fields, separated by a comma
 *  
 *  mgd_tree_build() may not be used outside of the lib.
 *  
 */
tree_node * mgd_tree_build(midgard * mgd, midgard_pool * pool,
				const char * table, const char * upfield, int root,
				int maxlevel, const char * sort)
{
	int root_ids[2] = {root, 0};
	int *child_ids = root_ids;
	tree_node *root_node, *parent_nodes, *child_nodes;
	int parent_num, child_num, i, j, k;

	if(sort && (!strncmp(sort, "id", 2) || !strncmp(sort, "up", 2)
						|| !strncmp(sort, "sitegroup", 9))) // who knows ?? ;)
		sort = NULL;
	ALLOC_TNODE(pool, root_node);
	root_node->id = root;
	parent_num = 1;
	parent_nodes = root_node;
	for(i = 0; !maxlevel || (i < maxlevel); i++) {
		child_num = mgd_tree_get_level(mgd, pool, table, upfield, sort,
									&child_ids, &child_nodes);
		if(!child_num) break;
		// if more parents than children, loop thru parents once
		if(parent_num > child_num) {
			for(j = 0; j < parent_num; j++) {
				for(k = 0; k < child_num; k++) {
					// zap child until we find the next one with parent
					if(parent_nodes[j].id != child_nodes[k].up) continue;
					parent_nodes[j].child = child_nodes + k;
					break;
				}
			}
		} else { // more children than parents, loop thru children once
			for(k = 0; k < child_num; k++) {
				for(j = 0; j < parent_num; j++) {
					// zap parent until we find the next one with children
					if(parent_nodes[j].id != child_nodes[k].up) continue;
					parent_nodes[j].child = child_nodes + k;
					// found it, zap siblings
					while(child_nodes[k].next) k++;
				}
			}
		}
		parent_nodes = child_nodes;
		parent_num = child_num;
	}
	mgd_tree_init(root_node, 0);
	return root_node;
}

/*! \brief Execute a function for each node of the tree.
 *  \param mgd			Midgard handle
 *  \param table		Identify the type of the node
 *  \param upfield		A string containing the name of the up field
 *  \param root			The node id root of the subtree
 *  \param maxlevel		The maximum recursion depth
 *  \param order		Specify when the function is executed for each node
 *  					during the walk: 0-before, 1-after
 *  \param xparam		A user defined variable passed to the user function
 *  \param func			A user defined function called for each node
 *  \param sort			The sort fields, separated by a comma
 *  
 *  \return				non-0 if the tree walk has been completed, 0 otherwise
 *  mgd_walk_table_tree() is used to execute a function for
 *  each node of the arborescence.
 *  
 */
MGD_API int mgd_walk_table_tree(midgard * mgd, const char *table, const char *upfield,
			int root, int maxlevel, int order, void *xparam,
			 midgard_userfunc func, const char * sort)
{
	midgard_pool *pool;
	tree_node *tree;
	int ret;

	pool = mgd_alloc_pool();

	tree = mgd_tree_build(mgd, pool, table, upfield, root, maxlevel, sort);

	if (!tree) {
		return 0;
	}
	/* Be aware of top level trees */
	ret = walktree(mgd, tree, order, 0, maxlevel, xparam, func);
	mgd_free_pool(pool);
	return ret;
}

/*! \brief Get a flat tree ids list.
 *  \param mgd			Midgard handle
 *  \param table		Identify the type of the node
 *  \param upfield		A string containing the name of the up field
 *  \param id			The node id root of the subtree
 *  \param maxlevel		The maximum recursion depth
 *  \param sort			The sort fields, separated by a comma
 *  
 *  mgd_tree() flattens the tree and returns a list of ids corresponding to
 *  each node of the arborescence.
 *  
 */
MGD_API int *mgd_tree(midgard * mgd, const char *table, const char *upfield,
				int id, int maxlevel, const char *sort)
{
	int *ids, i = 0;
	tree_node * tree;
	midgard_pool * pool;

	pool = mgd_alloc_pool();
	tree = mgd_tree_build(mgd, pool, table, upfield, id, maxlevel, sort);
	if(!tree) return NULL;

	ids = (int *)malloc((tree->size + 1) * sizeof(int));
	
	mgd_tree_get_ids(tree, ids, &i);
	ids[i] = 0;

	mgd_free_pool(pool);
	return ids;
}

/*! \brief Check if an node is descendant of another node.
 *  \param mgd		Midgard handle
 *  \param table	Identify the type of object
 *  \param upfield	A string containing the name of the up field
 *  \param root		The root node id
 *  \param id		The node id we want to check
 *  
 *  This function may be used to check the ascendance of a node in a tree.
 *  
 */
MGD_API int mgd_is_in_tree(midgard * mgd, const char * table,
				const char * upfield, int root, int id)
{
	midgard_res * res;
	midgard_pool * pool;
	int level=0;

	pool = mgd_alloc_pool();

	if(!table) return 0;
	if(!upfield || !(*upfield)) upfield = "up";
	while(root != id) {
#if HAVE_MIDGARD_SITEGROUPS
		res = mgd_query(mgd,
			"SELECT $s FROM $s WHERE id=$d AND (sitegroup in (0, $d) OR $d<>0)",
			upfield, table, id, mgd_sitegroup(mgd), mgd_isroot(mgd));
#else
		res = mgd_query(mgd,
			"SELECT $s FROM $s WHERE id=$d",
			upfield, table, id);
#endif
		if(!res || !mgd_rows(res)) {
			mgd_free_pool(pool);
			return 0;
		}
		mgd_fetch(res);
		id = mgd_sql2int(res, 0);
		level++;
		mgd_release(res);
	}
	mgd_free_pool(pool);
	if(root == id) return level+1;
	return 0;
}
